"""Metaclass module utility functions."""

from .. import utl

__all__ = (
    'parse_new_annotations',
    'parse_new_namespace',
    *utl.__all__
    )

from ... import core

from .. import cfg
from .. import exc
from .. import lib
from .. import typ

from .. utl import *


class Constants(cfg.Constants):
    """Constant values specific to this file."""


def parse_new_annotations(
    __namespace: dict[typ.AnyString, lib.t.Any],
    __annotations: typ.SnakeDict,
    __module: str,
    __slots: list[typ.string[typ.snake_case]],
    __fields: typ.DataClassFields,
    __base_fields: set[typ.string[typ.snake_case]]
    ) -> tuple[
        list[typ.string[typ.snake_case]],
        typ.DataClassFields
        ] | lib.Never:
    """
    Parse passed `__annotations` for valid `Fields` not otherwise \
    explicitly mapped in `__namespace`.

    Returns updated `__slots` and `__fields`.

    ---

    Passed `__namespace` will be mutated inplace, removing all key, \
    value pairs where keys overlap with those in `__annotations`.

    ---

    Raises a corresponding exception if class definition invalid.

    """

    for name, dtype in __annotations.items():
        if (
            typ.utl.check.is_wrapper_type(dtype)
            or (name in __fields and name not in __base_fields)
            ):
            continue
        elif (
            not utl.is_valid_keyword(name)
            and __module != Constants.FIELDS_MODULE
            ):
            raise exc.ReservedKeywordError(name)
        elif not typ.utl.check.is_field_type(dtype):
            raise exc.FieldAnnotationError(name, dtype)
        elif not core.strings.utl.is_snake_case_string(name):
            raise exc.IncorrectCasingError(tuple(__annotations))
        elif (
            (default := __namespace.pop(name, Constants.UNDEFINED))
            == Constants.UNDEFINED
            ):
            required = True
            default = None
        else:
            required = False

        if __module == Constants.FIELDS_MODULE:
            __fields[name] = {
                'name': name,
                'type': dtype,
                'default': default,
                'required': required
                }
        else:
            from .. import fields as fields_
            __fields[name] = fields_.Field(
                name=name,
                type=typ.utl.check.get_args(dtype)[0],
                default=default,
                required=required
                )
        __slots.append(name)

    return __slots, __fields


def parse_new_namespace(
    __namespace: dict[typ.AnyString, lib.t.Any],
    __annotations: typ.SnakeDict,
    __module: str,
    __slots: list[typ.string[typ.snake_case]],
    __fields: typ.DataClassFields,
    ) -> tuple[
        list[str],
        list[typ.string[typ.snake_case]],
        typ.DataClassFields
        ] | lib.Never:
    """
    Parse passed `__namespace` for valid `Fields`.

    Returns a `list[str]` of default values, as well as updated \
    `__slots` and `__fields`.

    ---

    Raises a corresponding exception if class definition invalid.

    """

    defaults: list[str] = []
    for name, default in __namespace.items():
        if (
            (is_snake_case := core.strings.utl.is_snake_case_string(name))
            and (
                utl.is_valid_keyword(name)
                or __module == Constants.FIELDS_MODULE
                )
            ):
            dtype = __annotations.get(name)
        elif not is_snake_case:
            raise exc.IncorrectCasingError(tuple(__namespace))
        else:
            raise exc.ReservedKeywordError(name)

        if typ.utl.check.is_wrapper_type(dtype):
            continue
        elif (
            (is_field := typ.utl.check.is_field(default))
            and typ.utl.check.is_field_type(dtype)
            ):
            default['name'] = name
            default['type'] = typ.utl.check.get_args(dtype)[0]
        elif (
            is_field_as_dict := (
                isinstance(default, dict)
                and len(default) > 0
                and all(
                    k in Constants.FIELD_KEYS
                    for k
                    in default
                    )
                and __module != Constants.FIELDS_MODULE
                )
            and typ.utl.check.is_field_type(dtype)
            ):
            default['name'] = name
            default['type'] = typ.utl.check.get_args(dtype)[0]
            from .. import fields as fields_
            __fields[name] = fields_.Field(default)
        elif (
            (is_field or is_field_as_dict)
            and dtype is None
            ):
            raise exc.MissingTypeAnnotation(name)
        elif (
            (is_field or is_field_as_dict)
            and not typ.utl.check.is_field_type(dtype)
            ):
            raise exc.FieldAnnotationError(name, dtype)

        if is_field:
            __fields[name] = default
            __slots.append(name)
            defaults.append(name)
        elif is_field_as_dict:
            from .. import fields as fields_
            __fields[name] = fields_.Field(default)
            __slots.append(name)
            defaults.append(name)

    return defaults, __slots, __fields



"""Type hinting utility functions."""

__all__ = (
    'collect_annotations',
    'resolve_type',
    )

from .. import cfg
from .. import lib

from . import check

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from .. import obj
    from .. import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""

    CACHED_ANNOTATIONS: 'dict[str, typ.AnyDict]' = {}
    """Local cache for typed object annotations."""


eval_type: lib.t.Callable[
    [
        'typ.AnyOrForwardRef',
        lib.t.Any,
        lib.t.Any,
        lib.t.Optional[frozenset]
        ],
    lib.t.Any
    ] = lib.t._eval_type  # type: ignore[attr-defined]
"""
Evaluate all `ForwardRef` in the given `type`.

---

For use of globalns and localns see the docstring for `get_type_hints()`.

`recursive_guard` is used to prevent infinite recursion with a recursive
`ForwardRef`.

"""


@lib.t.overload
def parse_ref_to_typ(
    ref: lib.t.ForwardRef,
    globalns: None,
    localns: 'typ.OptionalAnyDict'
    ) -> lib.t.ForwardRef: ...
@lib.t.overload
def parse_ref_to_typ(
    ref: lib.t.ForwardRef,
    globalns: 'typ.OptionalAnyDict',
    localns: 'typ.OptionalAnyDict'
    ) -> 'typ.AnyOrForwardRef': ...
def parse_ref_to_typ(
    ref: lib.t.ForwardRef,
    globalns: 'typ.OptionalAnyDict' = None,
    localns: 'typ.OptionalAnyDict' = None
    ) -> 'typ.AnyOrForwardRef':
    """Attempt to cast `ForwardRef` to `type`."""

    try:
        tp = eval_type(
            ref,
            globalns,
            localns,
            frozenset()
            )
    except NameError:
        return ref
    else:
        return tp


def parse_str_to_ref(
    typ_as_str: str,
    is_argument: bool,
    ) -> lib.t.ForwardRef:
    """Cast `str` to `ForwardRef`."""

    return lib.t.ForwardRef(
        typ_as_str,
        is_argument=is_argument,
        is_class=True
        )


@lib.t.overload
def resolve_type(
    typ_ref_or_str: 'typ.AnyType | typ.StrOrForwardRef',
    globalns: 'typ.AnyDict',
    localns: 'typ.AnyDict',
    is_argument: bool
    ) -> 'typ.AnyType | lib.t.Any': ...
@lib.t.overload
def resolve_type(
    typ_ref_or_str: 'typ.AnyType | typ.StrOrForwardRef',
    globalns: 'typ.OptionalAnyDict',
    localns: 'typ.OptionalAnyDict',
    is_argument: bool
    ) -> 'typ.AnyType | typ.AnyOrForwardRef': ...
@lib.t.overload
def resolve_type(
    typ_ref_or_str: 'typ.StrOrForwardRef',
    globalns: 'typ.OptionalAnyDict' = None,
    localns: 'typ.OptionalAnyDict' = None,
    is_argument: bool = False
    ) -> 'typ.AnyOrForwardRef': ...
def resolve_type(
    typ_ref_or_str: 'typ.AnyType | typ.StrOrForwardRef',
    globalns: 'typ.OptionalAnyDict' = None,
    localns: 'typ.OptionalAnyDict' = None,
    is_argument: bool = False
    ) -> 'typ.AnyType | typ.AnyOrForwardRef':
    """
    Attempt to resolve `str` or `ForwardRef` to `type`.

    ---

    Recursively resolves parameterized generics.

    """

    if isinstance(typ_ref_or_str, str):
        ref = parse_str_to_ref(typ_ref_or_str, is_argument)
        return resolve_type(ref, globalns, localns, is_argument)
    elif check.is_params_type(typ_ref_or_str):
        args = check.get_type_args(typ_ref_or_str)
        for arg in args:
            resolve_type(arg, globalns, localns, True)
        return typ_ref_or_str
    elif isinstance(typ_ref_or_str, lib.t.ForwardRef):
        typ_or_ref = parse_ref_to_typ(typ_ref_or_str, globalns, localns)
        if check.is_params_type(typ_or_ref):
            return resolve_type(typ_or_ref, globalns, localns, is_argument)
        else:
            return typ_or_ref
    else:
        return typ_ref_or_str


def _collect_annotations(
    __name: str,
    __annotations: 'typ.AnyDict',
    __bases: tuple[type, ...]
    ) -> 'typ.AnyDict':
    annotations: 'typ.AnyDict' = {}
    for _base in reversed(__bases):
        for __base in reversed(_base.__mro__):
            annotations |= getattr(__base, '__annotations__', {})
    annotations |= __annotations
    # Ensure any annotations hinted for TYPE_CHECKING removed.
    annotations.pop(Constants.__ANNOTATIONS__, None)
    annotations.pop(Constants.__DATACLASS_FIELDS__, None)
    annotations.pop(Constants.__DICT__, None)
    annotations.pop(Constants.__HERITAGE__, None)
    annotations.pop(Constants.FIELDS, None)
    annotations.pop(Constants.ENUMERATIONS, None)
    annotations.pop(Constants.HASH_FIELDS, None)
    Constants.CACHED_ANNOTATIONS[__name] = annotations
    return annotations


def collect_annotations(
    typed_obj: 'obj.SupportsAnnotations | type[obj.SupportsAnnotations]'
    ) -> 'typ.AnyDict':
    """
    Get all type annotations for `typed_obj`.

    ---

    Walks `__bases__` to collect all annotations.

    """

    obj_tp = typed_obj if isinstance(typed_obj, type) else type(typed_obj)

    if obj_tp.__name__ in Constants.CACHED_ANNOTATIONS:
        return Constants.CACHED_ANNOTATIONS[obj_tp.__name__]

    return _collect_annotations(
        obj_tp.__name__,
        getattr(obj_tp, '__annotations__', {}),
        obj_tp.__bases__
        )



"""CLI utility functions."""

__all__ = (
    'main',
    )

from . import obj


def main() -> None:  # pragma: no cover
    """
    Main CLI entrypoint.

    Commands follow the structure:

    `$ fqr {fgr_module_name} ...`

    """

    args = obj.root_parser.parse_args()
    kwargs = args._get_kwargs()
    as_args = [v for k, v in kwargs if k != 'func']
    args.func(*as_args)  # type: ignore[attr-defined]

    return None



"""Docs enumerations."""

from .. import core

__all__ = (
    'NoCrawlPath',
    'SiteMapChangeFreq',
    'SupportedTheme',
    *core.enm.__all__
    )

from .. core . enm import *

from . import lib


class SiteMapChangeFreq(lib.enum.Enum):
    """Valid site map change frequencies."""

    daily = 'daily'
    weekly = 'weekly'
    monthly = 'monthly'


class SupportedTheme(lib.enum.Enum):
    """Valid sphinx docs themes."""

    agogo = 'agogo'
    alabaster = 'alabaster'
    bizstyle = 'bizstyle'
    classic = 'classic'
    haiku = 'haiku'
    nature = 'nature'
    pyramid = 'pyramid'
    scrolls = 'scrolls'
    sphinxdoc = 'sphinxdoc'
    traditional = 'traditional'


class NoCrawlPath(lib.enum.Enum):
    """Paths to be Disallowed in generated robots.txt files."""

    search = 'Disallow: /search/'
    api = 'Disallow: /api/'
    builds = 'Disallow: /builds/'



import typing
import unittest

import fqr

from ... import mocking

from . import cfg


class Constants(cfg.Constants):
    """Constant values specific to unit tests in this file."""


class TestMeta(unittest.TestCase):
    """Fixture for testing Meta."""

    def setUp(self) -> None:
        self.mcs = fqr.objects.metas.Meta
        self.cls = mocking.Derivative
        self.field = fqr.Field(
            name='str_field',
            default='value',
            type_=str,
            )
        return super().setUp()

    def test_01_dict_functionality(self):
        """Test Meta __getitem__."""

        self.assertIsInstance(
            self.cls['str_field'],
            fqr.Field
            )

    def test_02_dict_functionality(self):
        """Test Meta __getitem__ with type input returns Field alias."""

        self.assertIsInstance(
            fqr.Field[str],
            fqr.objects.typ.Field
            )

    def test_03_dict_functionality(self):
        """Test Meta __getitem__ with type input raises KeyError if not Field."""

        self.assertRaises(KeyError, lambda: self.cls[str])

    def test_04_dict_functionality(self):
        """Test __setitem__ raises correct exc if value is not FieldType."""

        self.assertRaises(
            fqr.objects.exc.IncorrectTypeError,
            lambda: self.mcs.__setitem__(self.cls, self.field.name, 'value')
            )

    def test_05_dict_functionality(self):
        """Test __setitem__ raises correct exc if Field with invalid name."""

        self.assertRaises(
            fqr.objects.exc.InvalidFieldRedefinitionError,
            lambda: self.mcs.__setitem__(
                self.cls,
                self.field.name,
                fqr.Field(
                    name='_str_field',
                    default='value',
                    type_=str,
                    )
                )
            )

    def test_06_dict_functionality(self):
        """Test __setitem__ raises correct exc if Field has invalid type."""

        self.assertRaises(
            fqr.objects.exc.IncorrectTypeError,
            lambda: self.mcs.__setitem__(
                self.cls,
                self.field.name,
                fqr.Field(
                    name='str_field',
                    default='value',
                    type_=int,
                    )
                )
            )

    def test_07_dict_functionality(self):
        """Test __setitem__ raises correct exc if Field does not exist."""

        self.assertRaises(
            fqr.objects.exc.InvalidFieldAdditionError,
            lambda: self.mcs.__setitem__(
                self.cls,
                'field_that_does_not_exist',
                fqr.Field(
                    name='str_field',
                    default='value',
                    type_=str,
                    )
                )
            )

    def test_08_repr_functionality(self):
        """Test __repr__ is nice."""

        self.assertEqual(
            repr(self.cls),
            fqr.core.lib.json.dumps(
                dict(self.cls),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                )
            )

    def test_09_dict_keys(self):
        """Test keys."""

        self.assertListEqual(
            list(self.cls.keys()),
            list(self.cls.__dataclass_fields__)
            )

    def test_10_dict_setitem(self):
        """Test __setitem__ actually works."""

        self.cls[self.field.name] = self.field
        self.assertEqual(
            self.cls()[self.field.name],
            self.field.default
            )

    def test_11_iter(self):
        """Test __iter__."""

        self.assertTupleEqual(
            tuple(self.mcs.__iter__(self.cls)),
            self.cls.fields
            )

    def test_12_dict_functionality(self):
        """Test __setitem__ raises correct exc if invalid default."""

        self.assertRaises(
            fqr.objects.exc.IncorrectDefaultTypeError,
            lambda: self.mcs.__setitem__(
                self.cls,
                self.field.name,
                fqr.Field(
                    name='str_field',
                    default=4,
                    type_=str,
                    )
                )
            )

    def test_13_classvar_skip(self):
        """Test ClassVar annotation skip."""

        self.mcs(
            'ExcTest',
            (fqr.objects.objs.obj.ObjectBase, ),
            {
                '__annotations__': {
                    self.field.name: typing.ClassVar[str],
                    },
                '__module__': self.__module__
                }
            )
        self.assertTrue(True)

    def test_14_undefined_field_error(self):
        """Test cannot set undefined field."""

        self.assertRaises(
            fqr.objects.exc.InvalidFieldAdditionError,
            lambda: self.mcs.__setitem__(
                self.cls,
                'field_that_does_not_exist',
                fqr.Field(
                    name='field_that_does_not_exist',
                    default='value',
                    type_=str,
                    )
                )
            )

    def test_15_incorrect_type_error(self):
        """Test cannot `setattr` invalid type for `Field`."""

        self.assertRaises(
            fqr.objects.exc.IncorrectTypeError,
            lambda: self.mcs.__setattr__(self.cls, 'bool_field', 42)
            )

    def test_16_setattr(self):
        """Test `Meta.__setattr__()`."""

        new_field = fqr.Field(name='bool_field', type_=bool, default=True)
        self.mcs.__setattr__(self.cls, 'bool_field', new_field)
        self.assertEqual(self.cls.bool_field, new_field)

    def test_17_contains(self):
        """Test `Meta.__contains__()`."""

        self.assertTrue('bool_field' in self.cls)

    def test_18_invalid_definition_from_casing(self):
        """Test cannot define `Object` with invalid casing."""

        def _fn():
            class _InvalidObj(fqr.Object):
                valid_case_class_var: fqr.core.lib.t.ClassVar[str] = 'test'
                notSnakeCase: fqr.Field[str] = 'test'

        self.assertRaises(fqr.objects.exc.IncorrectCasingError, _fn)


class TestExceptions(unittest.TestCase):
    """Fixture for testing exceptions."""

    def setUp(self) -> None:
        self.mcs = fqr.objects.metas.Meta
        self.field = fqr.Field(
            name='str_field',
            default='value',
            type_=str,
            )
        return super().setUp()

    def test_01_no_reserved_fields(self):
        """Test cannot create with reserved keyword overwrites."""

        self.assertRaises(
            fqr.objects.exc.ReservedKeywordError,
            lambda: self.mcs(
                'ExcTest',
                (fqr.objects.objs.obj.ObjectBase, ),
                {
                    '__cache__': {},
                    '__module__': self.__module__,
                    '__dataclass_fields__': {}
                    }
                ),
            )

    def test_02_no_annotations(self):
        """Test cannot create without annotations."""

        self.assertRaises(
            fqr.objects.exc.MissingTypeAnnotation,
            lambda: self.mcs(
                'ExcTest',
                (fqr.objects.objs.obj.ObjectBase, ),
                {
                    self.field.name: self.field,
                    '__annotations__': {},
                    '__module__': self.__module__
                    }
                ),
            )

    def test_03_non_field_annotations(self):
        """Test cannot create without FieldType annotations."""

        self.assertRaises(
            fqr.objects.exc.FieldAnnotationError,
            lambda: self.mcs(
                'ExcTest',
                (fqr.objects.objs.obj.ObjectBase, ),
                {
                    self.field.name: self.field,
                    '__annotations__': {
                        self.field.name: str,
                        },
                    '__module__': self.__module__
                    }
                ),
            )

    def test_04_no_reserved_annotations(self):
        """Test cannot create with reserved keyword overwrites."""

        self.assertRaises(
            fqr.objects.exc.ReservedKeywordError,
            lambda: self.mcs(
                'ExcTest',
                (fqr.objects.objs.obj.ObjectBase, ),
                {
                    '__annotations__': {
                        '__dataclass_fields__': {},
                        },
                    '__module__': self.__module__
                    }
                ),
            )

    def test_05_non_field_annotations(self):
        """Test cannot create without FieldType annotations."""

        self.assertRaises(
            fqr.objects.exc.FieldAnnotationError,
            lambda: self.mcs(
                'ExcTest',
                (fqr.objects.objs.obj.ObjectBase, ),
                {
                    '__annotations__': {
                        self.field.name: str,
                        },
                    '__module__': self.__module__
                    }
                ),
            )

    def test_06_inconsistent_casing(self):
        """Test cannot create with inconsistent field casing."""

        self.assertRaises(
            fqr.objects.exc.IncorrectCasingError,
            lambda: self.mcs(
                'ExcTest',
                (fqr.objects.objs.obj.ObjectBase, ),
                {
                    '__annotations__': {
                        'string_field': fqr.Field[str],
                        'stringField': fqr.Field[str],
                        },
                    '__module__': self.__module__
                    }
                ),
            )



import pickle
import unittest

import fqr


class TestExceptions(unittest.TestCase):
    """Fixture for testing exceptions."""

    def setUp(self) -> None:
        return super().setUp()

    def test_01_serialization(self):
        """Test exc serializes correctly."""

        exc = fqr.loggers.exc.InvalidLogMessageTypeError(42)
        dump = pickle.dumps(exc)
        reloaded: fqr.core.typ.PackageExceptionType = pickle.loads(dump)
        self.assertTupleEqual(exc.args, reloaded.args)



import fqr


class Pet(fqr.Object):
    """A pet."""

    id_: fqr.Field[str]
    _alternate_id: fqr.Field[str]

    name: fqr.Field[str]
    type: fqr.Field[str]
    in_: fqr.Field[str]
    is_tail_wagging: fqr.Field[bool] = True



"""Query module."""

__all__ = (
    'AndQuery',
    'ContainsQueryCondition',
    'EqQueryCondition',
    'GeQueryCondition',
    'GtQueryCondition',
    'InvertQuery',
    'LeQueryCondition',
    'LtQueryCondition',
    'NeQueryCondition',
    'OrQuery',
    'Query',
    'QuerySortBy',
    'SimilarQueryCondition',
    )

from .. import cfg
from .. import enm
from .. import fields
from .. import lib
from .. import objs
from .. import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""


class QuerySortBy(objs.Object):
    """Simple specification for sorting Query results by field."""

    field: fields.Field[str]
    direction: fields.Field[str] = fields.Field(
        default=enm.SortDirection.asc.value,
        enum=enm.SortDirection,
        )


class Query(objs.Object):
    """
    Database agnostic query object.

    ---

    ### Usage

    Queries for `Objects` can be generated from their fields \
    using the following comparison operators:

    * `field_1_eq_filter = Object.field_1 == 'test_value_123'`
    * `field_1_ne_filter = Object.field_1 != 'test_value_123'`
    * `field_1_ge_filter = Object.field_1 >= 'test_value_123'`
    * `field_1_gt_filter = Object.field_1 > 'test_value_123'`
    * `field_1_le_filter = Object.field_1 <= 'test_value_123'`
    * `field_1_lt_filter = Object.field_1 < 'test_value_123'`

    And the following special operators:

    * `field_1_contains_filter = Object.field_1 << 'test_value_123'`
    * `field_1_similarity_filter = Object.field_1 % 'test_value_123'`
    * `field_1_similarity_filter_with_threshold = Object.field_1 % ('test_value_123', 0.8)`

    Queries may be chained together using the `&` and `|` bitwise \
    operators, corresponding to `and` and `or` clauses respectively.

    Additionally, the invert (`~`) operator may be prefixed to any \
    Query to match the opposite of any conditions specified \
    instead.

    Queries also support optional result limiting and sorting:

    * Result limits can be specified by setting the `limit` field.
    * Results can be sorted any number of times using the `+=` and `-=` \
    operators.

    ---

    ### Example

    ```py
    query: Query = (
        (
            (Object.integer_field >= 1)
            | (Object.string_field % ('test', 0.75))
            )
        & ~(Object.list_field << 'test')
        ) += 'string_field' -= 'integer_field'

    ```

    In the example above, the query would match any `Object` for which \
    the string `'test'` is `not` a member of `list_field` and for which \
    either the value for `integer_field` is greater than or equal to `1` \
    or the value for `string_field` is at least `75%` similar to `'test'`. \
    Results would then be sorted first in `ascending` order on `string_field`, \
    then in `descending` order on `integer_field`.

    """

    sorting: fields.Field[list[QuerySortBy]] = []
    limit: fields.Field[lib.t.Optional[int]] = None

    def __iadd__(self, field: str) -> lib.Self:
        """Add an ascending sort to the selected `Query`."""

        self._sort_by(field)
        return self

    def __isub__(self, field: str) -> lib.Self:  # type: ignore[misc, override]
        """Add a descending sort to the selected `Query`."""

        self._sort_by(field, direction=enm.SortDirection.desc.value)
        return self

    def __and__(self, other: 'Query') -> 'AndQuery':
        return AndQuery(and_=[self, other])

    def __or__(self, other: 'Query') -> 'OrQuery':
        return OrQuery(or_=[self, other])

    def __invert__(self) -> 'InvertQuery':
        return InvertQuery(invert=self)

    def _sort_by(
        self,
        field: str,
        direction: typ.SortDirection = enm.SortDirection.asc.value
        ) -> None:
        """
        Sort the selected `Query` by `field` and `direction`.

        Direction may be either `'asc'` or `'desc'`.

        The `field` should correspond to an existing field for the `Object`.

        """

        self.sorting.append(QuerySortBy(field=field, direction=direction))


class QueryCondition(Query):
    """Base query filter."""

    field: fields.Field[str]


class SimilarQueryCondition(QueryCondition):
    """Filters where field is similar to value."""

    like: fields.Field[typ.Primitive | lib.t.Any]
    threshold: fields.Field[lib.t.Optional[float]] = None


class ContainsQueryCondition(QueryCondition):
    """Filters where field contains value."""

    contains: fields.Field[typ.Primitive | lib.t.Any]


class EqQueryCondition(QueryCondition):
    """Filters where == value."""

    eq: fields.Field[typ.Primitive | lib.t.Any]


class NeQueryCondition(QueryCondition):
    """Filters where != value."""

    ne: fields.Field[typ.Primitive | lib.t.Any]


class LeQueryCondition(QueryCondition):
    """Filters where <= value."""

    le: fields.Field[typ.Primitive | lib.t.Any]


class LtQueryCondition(QueryCondition):
    """Filters where < value."""

    lt: fields.Field[typ.Primitive | lib.t.Any]


class GeQueryCondition(QueryCondition):
    """Filters where >= value."""

    ge: fields.Field[typ.Primitive | lib.t.Any]


class GtQueryCondition(QueryCondition):
    """Filters where > value."""

    gt: fields.Field[typ.Primitive | lib.t.Any]


class AndQuery(Query):
    """Filters on all conditions."""

    and_: fields.Field[list[Query]]


class OrQuery(Query):
    """Filters on any condition."""

    or_: fields.Field[list[Query]]


class InvertQuery(Query):
    """Inverts the filter."""

    invert: fields.Field[Query]



"""
Overview
========

**Summary:** Primary command-line interface (CLI) to fqr.

---

Usage
-----

```sh
$ fqr --help
```

"""

__all__ = (
    'cfg',
    'lib',
    'obj',
    'main',
    'utl',
    )

from . import cfg
from . import lib
from . import obj
from . import utl

from . utl import main



"""Loggers typing."""

from .. import core

__all__ = (
    'AnyLogRecord',
    'LogRecordType',
    'LogRecordWithPrint',
    'LogRecordWithPrintAndTraceBack',
    'LogRecordWithTraceBack',
    'SupportedLogObject',
    *core.typ.__all__
    )

from .. core . typ import *

from . import lib

AnyLogRecord: lib.t.TypeAlias = lib.t.Union[
    'LogRecord',
    'LogRecordWithPrint',
    'LogRecordWithPrintAndTraceBack',
    'LogRecordWithTraceBack',
    ]
SupportedLogObject: lib.t.TypeAlias = lib.t.Union[
    'Mapping[AnyString, SupportedLogObject]',
    Object,
    type[Object],
    AnyString,
    'Array[SupportedLogObject]'
    ]


class LogRecord(lib.t.TypedDict):
    """Basic log record."""

    message: SupportedLogObject


class LogRecordWithPrint(LogRecord):
    """Log record with print message included."""

    printed: str


class LogRecordWithTraceBack(LogRecord):
    """Log record with traceback included."""

    traceback: str


class LogRecordWithPrintAndTraceBack(LogRecord):
    """A fully populated log record."""

    printed: str
    traceback: str


LogRecordType = lib.t.TypeVar('LogRecordType', bound=AnyLogRecord)



"""Test commit msg validation."""

import unittest
import re

from . import cfg


class Constants(cfg.Constants):
    """Module specific constant values."""

    ValidMsgPattern = re.compile(
        r'^([mM]erge .*)$'
        '|'
        '(^'
        r'((build|chore|ci|docs|feat|fix|perf|refactor|revert|style|test)!?: .+)'
        '('
        r'(\n\n(.+)\n\n)'
        r'((BREAKING CHANGE|DEPRECATED)(: )(.+)\n\n(.+)\n\n\n)?'
        r'(resolve[ds]? \#[A-Z0-9\-]+|fix(ed|es)? \#[A-Z0-9\-]+|close[ds]? \#[A-Z0-9\-]+)'
        r'((, )(resolve[ds]? \#[A-Z0-9\-]+|fix(ed|es)? \#[A-Z0-9\-]+|close[ds]? \#[A-Z0-9\-]+))?'
        ')?'
        '$)'
        '|'
        '(^'
        r'revert: ((build|chore|ci|docs|feat|fix|perf|refactor|revert|style|test)!?: .+)'
        r'(\n\n(This reverts commit [a-z0-9]{40}\..*)\n\n)'
        r'(fix(ed|es)? \#[A-Z0-9\-]+)'
        r'((, )(fix(ed|es)? \#[A-Z0-9\-]+))?'
        '$)'
        )


class TestMessageValidation(unittest.TestCase):
    """Tests for commit msg validation."""

    def test_01_simple_msg(self):
        """Test short version works."""

        self.assertTrue(
            bool(
                Constants
                .ValidMsgPattern
                .match(
                    'test: __short_valid_example__'
                    )
                )
            )

    def test_02_regular_msg(self):
        """Test normal version works."""

        self.assertTrue(
            bool(
                Constants
                .ValidMsgPattern
                .match(
                    'feat: __valid_example__'
                    '\n\n'
                    'optional body text'
                    '\n\n'
                    'closes #1, resolve #2'
                    )
                )
            )

    def test_03_breaking_changes_msg(self):
        """Test breaking changes version works."""

        self.assertTrue(
            bool(
                Constants
                .ValidMsgPattern
                .match(
                    'feat!: __new_stuff__'
                    '\n\n'
                    'body text.'
                    '\n\n'
                    'BREAKING CHANGE: Breaks stuff.'
                    '\n\n'
                    'Details on how stuff breaks and what to do.'
                    '\n\n\n'
                    'resolves #1'
                    )
                )
            )

    def test_04_revert_msg(self):
        """Test revert version works."""

        self.assertTrue(
            bool(
                Constants
                .ValidMsgPattern
                .match(
                    'revert: feat!: __new_stuff__'
                    '\n\n'
                    'This reverts commit'
                    ' 2c4ed28b069267f39974b5da50795c5210040e33. Because'
                    ' reasons.'
                    '\n\n'
                    'fixes #TKT-123'
                    )
                )
            )



"""Module utils unit tests."""

import unittest

import fqr

from fqr . core import codecs
from fqr . core import lib

from . import cfg


class SimpleTypedObj(lib.t.TypedDict):
    """Simple `SupportsAnnotation` object."""

    name: str
    id_: int


class SubDecimal(lib.decimal.Decimal):
    """Decimal subclass for testing."""


class UnknownSerializable:
    """Class for testing with no known serialization."""


class Constants(cfg.Constants):
    """Constant values specific to unit tests in this file."""

    SimpleDict: dict[str, tuple[int, ...]] = {
        'a_simple_key': (0, ),
        '_and_another_': (4, 3, 2, ),
        'components': (1, 2, )
        }
    SimpleTuple = (1, 2, 3, )
    BoolTuple = (True, False, True, )
    BoundType = lib.t.TypeVar('BoundType', bound=int)
    ComplexStr = '1.134_12e+2-1.134_12e+2j'
    ConstrainedType = lib.t.TypeVar('ConstrainedType', bool, int)
    AnotherConstrainedType = lib.t.TypeVar(
        'AnotherConstrainedType',
        tuple[int] | tuple[str],
        tuple[bool] | tuple[int],
        tuple[float] | tuple[bool] | tuple[int] | bool
        )
    NestedDict = {'nesting': SimpleDict}
    DateTime = lib.datetime.datetime.now(lib.datetime.timezone.utc)
    Date = DateTime.date()
    SimpleObj = SimpleTypedObj(name='test', id_=1)


class TestUtils(unittest.TestCase):
    """Fixture for testing."""

    def test_01_tuple_parse(self):
        """Test `parse` on `tuple[int, ...]`."""

        self.assertEqual(
            Constants.SimpleTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                tuple[int, ...]
                )
            )

    def test_02_dict_parse(self):
        """Test `parse` on `dict[str, tuple[int, ...]]`."""

        self.assertEqual(
            Constants.SimpleDict,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleDict),
                dict[str, tuple[int, ...]]
                )
            )

    def test_03_typevar_parse_bound(self):
        """Test `parse` on `tuple[BoundType, ...]`."""

        self.assertNotEqual(
            Constants.SimpleTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.BoolTuple),
                tuple[Constants.BoundType, ...]
                )
            )

    def test_04_typevar_parse_bound(self):
        """Test `parse` on `tuple[BoundType, ...]`."""

        self.assertEqual(
            Constants.BoolTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.BoolTuple),
                tuple[Constants.BoundType, ...]
                )
            )

    def test_05_typevar_parse_constrained(self):
        """Test `parse` on `tuple[ConstrainedType, ...]`."""

        self.assertEqual(
            Constants.BoolTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.BoolTuple),
                tuple[Constants.ConstrainedType, ...]
                )
            )

    def test_06_nested_dict_parse(self):
        """Test `parse` on `dict[str, dict[str, tuple[int, ...]]]`."""

        self.assertEqual(
            Constants.NestedDict,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.NestedDict),
                dict[str, dict[str, tuple[int, ...]]]
                )
            )

    def test_07_bool_parse(self):
        """Test `parse` on `bool`."""

        self.assertIs(True, codecs.utl.parse('true', bool))

    def test_08_anti_bool_parse(self):
        """Test `parse` on `bool`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.bool_decode,
            codecs.utl.parse('asdf', bool)
            )

    def test_09_float_parse(self):
        """Test `parse` on `float`."""

        self.assertEqual(1.8, codecs.utl.parse('1.8', float))

    def test_10_anti_float_parse(self):
        """Test `parse` on `float`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.number_decode,
            codecs.utl.parse('asdf', float)
            )

    def test_11_complex_parse(self):
        """Test `parse` on `complex`."""

        self.assertEqual(
            complex(Constants.ComplexStr),
            codecs.utl.parse(Constants.ComplexStr, complex)
            )

    def test_12_none_parse(self):
        """Test `parse` on `None`."""

        self.assertIsNone(codecs.utl.parse('null', type(None)))

    def test_13_anti_none_parse(self):
        """Test `parse` on `None`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.null_decode,
            codecs.utl.parse('asdf', type(None))
            )

    def test_14_union_parse(self):
        """Test `parse` on `int | str`."""

        self.assertIsInstance(codecs.utl.parse('42', tuple[int] | int), int)

    def test_15_anti_tuple_parse(self):
        """Test `parse` on `tuple[int, ...]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_arr_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                tuple[frozenset, ...]
                )
            )

    def test_16_known_tuple_parse(self):
        """Test `parse` on `tuple[int, int, int]`."""

        self.assertEqual(
            Constants.SimpleTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                tuple[int, int, int]
                )
            )

    def test_17_anti_known_tuple_parse(self):
        """Test `parse` on `tuple[int, int, int]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_arr_len,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                tuple[int, int]
                )
            )

    def test_18_anti_known_tuple_parse(self):
        """Test `parse` on `tuple[int, int, int]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_arr_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                tuple[frozenset, list, tuple]
                )
            )

    def test_19_list_parse(self):
        """Test `parse` on `list[int]`."""

        self.assertEqual(
            list(Constants.SimpleTuple),
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                list[int]
                )
            )

    def test_20_anti_list_parse(self):
        """Test `parse` on `list[int]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.value_decode,
            codecs.utl.parse('42', list[int])
            )

    def test_21_anti_list_parse(self):
        """Test `parse` on `list[int]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_arr_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                list[frozenset]
                )
            )

    def test_22_anti_list_parse(self):
        """Test `parse` on `list[int]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_json,
            codecs.utl.parse('asdf', list[frozenset])
            )

    def test_23_anti_dict_parse(self):
        """Test `parse` on `dict[str, tuple[int, ...]]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_json,
            codecs.utl.parse('asdf', dict[str, tuple[int, ...]])
            )

    def test_24_anti_dict_parse(self):
        """Test `parse` on `dict[str, tuple[int, ...]]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_keys_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(
                    {
                        i: v
                        for (i, v)
                        in enumerate(Constants.SimpleDict.values())
                        }
                    ),
                dict[tuple[str], tuple[int, ...]]
                )
            )

    def test_25_anti_dict_parse(self):
        """Test `parse` on `dict[str, tuple[int, ...]]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_values_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(
                    {
                        k: i
                        for (i, k)
                        in enumerate(Constants.SimpleDict.keys())
                        }
                    ),
                dict[str, tuple[int, ...]]
                )
            )

    def test_26_anti_dict_parse(self):
        """Test `parse` on `dict[str, tuple[int, ...]]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_map_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleDict),
                dict[str]
                )
            )

    def test_27_anti_dict_parse(self):
        """Test `parse` on `dict[str, tuple[int, ...]]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.value_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                dict[str, tuple[int, ...]]
                )
            )

    def test_28_typevar_parse_unbound(self):
        """Test `parse` on `tuple[BoundType, ...]`."""

        self.assertEqual(
            Constants.BoolTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.BoolTuple),
                tuple[fqr.core.typ.AnyType, ...]
                )
            )

    def test_29_typevar_parse_constrained_again(self):
        """Test `parse` on `tuple[AnotherConstrainedType, ...]`."""

        self.assertEqual(
            Constants.BoolTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.BoolTuple),
                tuple[Constants.AnotherConstrainedType, ...]
                )
            )

    def test_30_datetime_parse(self):
        """Test `parse` on `datetime.datetime`."""

        self.assertEqual(
            Constants.DateTime,
            codecs.utl.parse(
                Constants.DateTime.isoformat(),
                lib.datetime.datetime
                )
            )

    def test_31_date_parse(self):
        """Test `parse` on `datetime.date`."""

        self.assertEqual(
            Constants.Date,
            codecs.utl.parse(
                Constants.Date.isoformat(),
                lib.datetime.date
                )
            )

    def test_32_anti_datetime_parse(self):
        """Test `parse` on `datetime.datetime`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.datetime_decode,
            codecs.utl.parse('123', lib.datetime.datetime)
            )

    def test_33_parse_literal(self):
        """Test `parse` on `tuple[Literal[1], Literal[2], Literal[3]]`."""

        self.assertEqual(
            Constants.SimpleTuple,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                tuple[
                    lib.t.Literal[1],
                    lib.t.Literal[2],
                    lib.t.Literal[3]
                    ]
                )
            )

    def test_34_anti_parse_literal(self):
        """Test `parse` on `tuple[Literal[1], Literal[2], Literal[3]]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_arr_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleTuple),
                tuple[
                    lib.t.Literal[3],
                    lib.t.Literal[2],
                    lib.t.Literal[1]
                    ]
                )
            )

    def test_35_parse_another_literal(self):
        """Test `parse` on `Literal[1]`."""

        self.assertEqual(
            1,
            codecs.utl.parse('1', lib.t.Literal[1])
            )

    def test_36_anti_parse_another_literal(self):
        """Test `parse` on `Literal[1]`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.literal_decode,
            codecs.utl.parse('2', lib.t.Literal[1])
            )

    def test_37_parse_typed_obj(self):
        """Test `parse` on `TypedDict`."""

        self.assertEqual(
            Constants.SimpleObj,
            codecs.utl.parse(
                codecs.lib.json.dumps(Constants.SimpleObj),
                SimpleTypedObj
                )
            )

    def test_38_anti_parse_typed_obj(self):
        """Test `parse` on `TypedDict`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_map_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(SimpleTypedObj(name=1, id_='test')),
                SimpleTypedObj
                )
            )

    def test_39_anti_parse_typed_obj(self):
        """Test `parse` on `TypedDict`."""

        self.assertEqual(
            codecs.enm.ParseErrorRef.invalid_keys_decode,
            codecs.utl.parse(
                codecs.lib.json.dumps(SimpleTypedObj(not_a_key='test')),
                SimpleTypedObj
                )
            )

    def test_40_encode_subclass(self):
        """Test `encode` on `SubDecimal(Decimal)`."""

        value = SubDecimal(0.1)
        self.assertEqual(float(value), codecs.utl.encode(value))

    def test_41_encode_unknown(self):
        """Test `encode` on `UnknownSerializable`."""

        self.assertEqual(
            repr(UnknownSerializable),
            codecs.utl.encode(UnknownSerializable)
            )



"""Objects exceptions module."""

from .. import core

__all__ = (
    'FieldAnnotationError',
    'IncorrectCasingError',
    'IncorrectDefaultTypeError',
    'IncorrectTypeError',
    'InvalidComparisonTypeError',
    'InvalidContainerComparisonTypeError',
    'InvalidFieldAdditionError',
    'InvalidFieldRedefinitionError',
    'InvalidObjectComparisonError',
    'MissingTypeAnnotation',
    'ReservedKeywordError',
    'TypeValidationError',
    *core.exc.__all__
    )

from .. core . exc import *

from . import cfg
from . import lib


class Constants(cfg.Constants):
    """Constant values specific to this file."""


class InvalidFieldAdditionError(BasePackageException[str]):
    """Cannot add new fields after a class has already been defined."""

    def __init__(self, name: str):
        super().__init__(
            ' '.join(
                (
                    'Cannot add fields after a class has already',
                    'been defined.',
                    f'\nFIELD: {name}'
                    )
                ),
            name
            )


class InvalidFieldRedefinitionError(BasePackageException[str]):
    """Cannot redefine field for the same key with a different name."""

    def __init__(self, name: str):
        super().__init__(
            ' '.join(
                (
                    'Cannot redefine field for the same key',
                    'with a different name.',
                    f'\nFIELD: {name}'
                    )
                ),
            name
            )


class IncorrectCasingError(BasePackageException[lib.t.Iterable[str]]):
    """Incorrect field casing."""

    def __init__(self, fields: lib.t.Iterable[str]):
        super().__init__(
            ' '.join(
                (
                    'All fields for all Object derivatives',
                    'must be `string[snake_case]`.',
                    f'\nFIELDS: {sorted(fields)!s}',
                    )
                ),
            fields
            )


class IncorrectDefaultTypeError(
    BasePackageException[str, type[lib.t.Any], lib.t.Any]
    ):
    """
    Error raised when a field's default value is not of an allowed type.

    """

    def __init__(
        self,
        name: str,
        dtype: type[lib.t.Any],
        value: lib.t.Any
        ) -> None:
        super().__init__(
            ' '.join(
                (
                    f"Field: '{name}',",
                    f"only supports type: '{dtype!s}',",
                    f'Default supplied: {value!s}',
                    f"is of type: '{type(value)!s}'",
                    )
                ),
            *(name, dtype, value)
            )


class IncorrectTypeError(
    BasePackageException[str, lib.t.Any, lib.t.Any]
    ):
    """Error raised when a field value is not of an allowed type."""

    def __init__(
        self,
        name: str,
        dtype: lib.t.Any,
        value: lib.t.Any
        ) -> None:
        super().__init__(
            ' '.join(
                (
                    f"Field: '{name}',",
                    f"only supports type: '{dtype!s}',",
                    f'Value supplied: {value!s}',
                    f"is of type: '{type(value)!s}'",
                    )
                ),
            *(name, dtype, value)
            )


class InvalidComparisonTypeError(
    BasePackageException[str, lib.t.Any, lib.t.Any]
    ):
    """
    Error raised when comparing a field with a value of a different \
    type.

    """

    def __init__(
        self,
        name: str,
        dtype: lib.t.Any,
        value: lib.t.Any
        ) -> None:
        super().__init__(
            ' '.join(
                (
                    f"Cannot compare field: '{name}',",
                    f"of type: '{dtype!s}',",
                    f'with the value supplied: {value!s}',
                    f"of type: '{type(value)!s}'",
                    )
                ),
            *(name, dtype, value)
            )


class InvalidContainerComparisonTypeError(
    BasePackageException[str, lib.t.Any, lib.t.Any]
    ):
    """
    Error raised when checking for membership or similarity against a \
    non-Iterable field.

    """

    def __init__(
        self,
        name: str,
        dtype: lib.t.Any,
        value: lib.t.Any
        ) -> None:
        super().__init__(
            ' '.join(
                (
                    f"Field: '{name}',",
                    f"of type: '{dtype!s}',",
                    f'cannot contain or be similar to: {value!s},',
                    'as this field is not an iterable.'
                    )
                ),
            *(name, dtype, value)
            )


class InvalidObjectComparisonError(
    BasePackageException[lib.t.Any, lib.t.Any]
    ):
    """
    Error raised when comparing an object with another with differing \
    fields.

    """

    def __init__(
        self,
        obj: lib.t.Any,
        other: lib.t.Any
        ) -> None:
        super().__init__(
            ' '.join(
                (
                    f"Cannot compare object: '{obj!s}',",
                    f"of type: '{obj.__class__.__name__}',",
                    f'with object: {other!s}',
                    f"of type: '{other.__class__.__name__}'",
                    )
                ),
            *(obj, other)
            )


class FieldAnnotationError(
    BasePackageException[
        str,
        lib.t.Optional[lib.t.Union[type[lib.t.Any], lib.t.ForwardRef]]
        ]
    ):
    """Incomplete type annotation."""

    def __init__(
        self,
        name: str,
        dtype: lib.t.Optional[
            lib.t.Union[
                type[lib.t.Any],
                lib.t.ForwardRef
                ]
            ]
        ):
        stype: str = getattr(dtype, '__name__', str(dtype))
        super().__init__(
            ' '.join(
                (
                    'All type annotations for Object derivatives',
                    'must be of a generic Field type such as',
                    '`Field[int] or Field[str]`.',
                    f'\nFIELD: {name}, TYPE: {stype}',
                    f'\nSUGGESTION: Field[{stype}]',
                    )
                ),
            *(name, dtype)
            )


class MissingTypeAnnotation(BasePackageException[str]):
    """Incomplete type annotation."""

    def __init__(self, name: str):
        super().__init__(
            ' '.join(
                (
                    'Type for all Fields must be annotated.',
                    f'\nFIELD: {name}',
                    )
                ),
            name
            )


class ReservedKeywordError(BasePackageException[str]):
    """Invalid keyword."""

    def __init__(self, name: str):
        super().__init__(
            ' '.join(
                (
                    'The following keyword is reserved for special',
                    f'purposes within {Constants.PACAKGE} and may',
                    'not be used / overwritten in class definitions.',
                    f'\nKEYWORD: {name}',
                    f'\nSUGGESTION: {name}_',
                    )
                ),
            name
            )


class TypeValidationError(
    BasePackageException[str, lib.t.Any, core.codecs.enm.ParseErrorRef]
    ):
    """Error raised when a field value could not be parsed as valid type."""

    def __init__(
        self,
        name: str,
        dtype: lib.t.Any,
        error_ref: core.codecs.enm.ParseErrorRef
        ) -> None:
        super().__init__(
            ' '.join(
                (
                    error_ref.value,
                    f"Field: '{name}',",
                    f"only supports type: '{dtype!s}',"
                    )
                ),
            *(name, dtype, error_ref)
            )



"""Loggers utility functions."""

__all__ = (
    'parse_incoming_log_message',
    )

from .. import core

from . import cfg
from . import exc
from . import lib
from . import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""


@lib.t.overload
def parse_incoming_log_message(
    msg: str,
    level: lib.t.Literal[30]
    ) -> typ.LogRecord | typ.LogRecordWithPrint: ...
@lib.t.overload
def parse_incoming_log_message(
    msg: str,
    level: (
        lib.t.Literal[0]
        | lib.t.Literal[10]
        | lib.t.Literal[20]
        | lib.t.Literal[40]
        | lib.t.Literal[50]
        )
    ) -> typ.LogRecord: ...
@lib.t.overload
def parse_incoming_log_message(
    msg: lib.t.Any,
    level: lib.t.Literal[30]
    ) -> typ.LogRecord | typ.LogRecordWithPrint | lib.Never: ...
@lib.t.overload
def parse_incoming_log_message(
    msg: lib.t.Any,
    level: (
        lib.t.Literal[0]
        | lib.t.Literal[10]
        | lib.t.Literal[20]
        | lib.t.Literal[40]
        | lib.t.Literal[50]
        )
    ) -> typ.LogRecord | lib.Never: ...
def parse_incoming_log_message(
    msg: lib.t.Any,
    level: (
        lib.t.Literal[0]
        | lib.t.Literal[10]
        | lib.t.Literal[20]
        | lib.t.Literal[30]
        | lib.t.Literal[40]
        | lib.t.Literal[50]
        )
    ) -> typ.LogRecord | typ.LogRecordWithPrint | lib.Never:
    """
    Parse incoming log message or warning to dict format.

    ---

    Raises an exception if msg type cannot be parsed.

    Automatically redacts common, sensitive data.

    """

    if isinstance(msg, str):
        if level != lib.logging.WARNING:
            return typ.LogRecord(message=msg)
        else:
            original_msg, _, warn_msg = msg.partition('.py')
            if warn_msg:
                warn_msg = ':'.join(warn_msg.split(':')[3:]).strip()
                warn_msg, *_ = warn_msg.rpartition('\n')
                if Constants.SILENCE_MSG in warn_msg:
                    warn_msg, _, printed = warn_msg.partition('\n')
                    return typ.LogRecordWithPrint(
                        message=warn_msg,
                        printed=printed
                        )
                else:
                    return typ.LogRecord(message=warn_msg)
            else:
                return typ.LogRecord(message=original_msg)
    elif core.typ.utl.check.is_object(msg):
        if isinstance(msg, type):
            return typ.LogRecord(message={msg.__name__: msg})
        else:
            return typ.LogRecord(message={msg.__class__.__name__: msg})
    elif core.typ.utl.check.is_array(msg):
        return typ.LogRecord(message=msg)
    elif core.typ.utl.check.is_mapping(msg):
        if (
            (dict_keys := sorted(msg.keys())) == ['message']
            or dict_keys == ['message', 'printed']
            ):
            msg_: typ.LogRecord | typ.LogRecordWithPrint = msg
            return msg_
        else:
            return typ.LogRecord(message=msg)
    else:
        raise exc.InvalidLogMessageTypeError(msg)



"""
Overview
========

**Summary:** fqr extension for generating wiki documentation with [Sphinx](https://www.sphinx-doc.org).

---

Usage
-----

##### Document a package as follows.

```sh
$ pip install fqr[docs]
$ fqr docs package_name
```

"""

__all__ = (
    'cfg',
    'enm',
    'lib',
    'obj',
    'utl',
    )

from . import cfg
from . import enm
from . import lib
from . import obj
from . import utl



"""Objects constants."""

__all__ = (
    'Constants',
    )

from .. import core


class Constants(core.cfg.Constants):
    """Constant values specific to objects modules."""

    BASE_ATTRS = (
        '__heritage__',
        '__dataclass_fields__',
        'enumerations',
        'fields',
        'hash_fields',
        )
    FIELD_KEYS = (
        'name',
        'type',
        'default',
        'required',
        'enum',
        'min_length',
        'max_length',
        'minimum',
        'exclusive_minimum',
        'maximum',
        'exclusive_maximum',
        'multiple_of',
        'pattern',
        'min_items',
        'max_items',
        'unique_items',
        'read_only',
        'write_only',
        )
    FIELDS_MODULE = __name__.replace('cfg', 'fields.obj')
    OBJECTS_MODULE = __name__.replace('cfg', 'objs.obj')
    FORBIDDEN_KEYWORDS = (
        '__init__',
        '__init_subclass__',
        '__new__',
        '__hash__',
        '__getattribute__',
        '__bool__',
        '__contains__',
        '__getitem__',
        '__setitem__',
        '__delitem__',
        '__reversed__',
        '__eq__',
        '__ne__',
        '__iter__',
        '__ior__',
        '__len__',
        '__repr__',
        '__sub__',
        '__lshift__',
        '__rshift__',
        '__getstate__',
        '__setstate__',
        '__instancecheck__',
        '__subclasscheck__',
        'class_as_dict',
        'copy',
        'enumerations',
        'fields',
        'get',
        'hash_fields',
        'items',
        'keys',
        'pop',
        'setdefault',
        'update',
        'values',
        )



"""Core constants."""

__all__ = (
    'Constants',
    )

from . import lib

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from . import typ


class Constants:
    """Constant values shared across all of fqr."""

    PACAKGE = 'fqr'
    """Reference to package name."""

    CLASS_AS_DICT: 'typ.string[typ.snake_case]' = 'class_as_dict'
    """Common string reference."""

    DELIM = '-'
    """Common delimiter used throughout package."""

    INDENT = int(lib.os.getenv('INDENT', 2))
    """
    Default, package-wide indentation.

    ---

    Used for `__repr__`, log messages, etc.

    """

    UNDEFINED = f'[[{PACAKGE.upper()}_DEFAULT_PLACEHOLDER]]'
    """Placeholder for undefined values that should not be `None`."""

    ENV = lib.os.getenv('ENV', 'local').lower()
    """The lowered name of our runtime environment."""

    DEPLOY_ENVS = (
        'dev',
        'develop',
        'qa',
        'test',
        'testing',
        'stg',
        'stage',
        'staging',
        'uat',
        'prod',
        'production',
        )
    """Valid, non-local runtime environment names."""

    DELIM_REBASE = '_X_'

    __ANNOTATIONS__: 'typ.string[typ.snake_case]' = '__annotations__'
    __DATACLASS_FIELDS__: 'typ.string[typ.snake_case]' = '__dataclass_fields__'
    __DICT__: 'typ.string[typ.snake_case]' = '__dict__'
    __HERITAGE__: 'typ.string[typ.snake_case]' = '__heritage__'
    __SLOTS__: 'typ.string[typ.snake_case]' = '__slots__'
    __MODULE__: 'typ.string[typ.snake_case]' = '__module__'

    FIELDS: 'typ.string[typ.snake_case]' = 'fields'
    ENUMERATIONS: 'typ.string[typ.snake_case]' = 'enumerations'
    HASH_FIELDS: 'typ.string[typ.snake_case]' = 'hash_fields'



"""Codec constants."""

__all__ = (
    'Constants',
    )

from .. import cfg

from . import lib
from . import typ


def _isoformat_encoder(
    o: lib.datetime.datetime | lib.datetime.date | lib.datetime.time
    ) -> str:
    return o.isoformat()


class Constants(cfg.Constants):
    """Constant values shared across core codecs modules."""

    ENCODERS: dict[
        type[lib.t.Any],
        lib.t.Callable[[lib.t.Any], typ.Serial]
        ] = {
            bytes: lambda o: getattr(o, 'decode')(),
            lib.datetime.date: _isoformat_encoder,
            lib.datetime.datetime: _isoformat_encoder,
            lib.datetime.time: _isoformat_encoder,
            lib.datetime.timedelta: lambda td: getattr(td, 'total_seconds')(),
            lib.decimal.Decimal: float,
            lib.enum.Enum: lambda o: getattr(o, 'value'),
            frozenset: list,
            lib.collections.deque: list,
            lib.types.GeneratorType: list,
            lib.ipaddress.IPv4Address: str,
            lib.ipaddress.IPv4Interface: str,
            lib.ipaddress.IPv4Network: str,
            lib.ipaddress.IPv6Address: str,
            lib.ipaddress.IPv6Interface: str,
            lib.ipaddress.IPv6Network: str,
            lib.pathlib.Path: str,
            lib.re.Pattern: lambda o: getattr(o, 'pattern'),
            set: list,
            lib.uuid.UUID: str,
            }



"""Objects typing."""

from .. import core

__all__ = (
    'Field',
    'MetaType',
    'SortDirection',
    'Type',
    *core.typ.__all__,
    )

from .. core . typ import *

from . import lib

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from . import metas  # noqa: F401

SortDirection = lib.t.Literal['asc'] | lib.t.Literal['desc']

MetaType = lib.t.TypeVar('MetaType', bound='metas.Meta')
Type = lib.t.TypeVar('Type', bound=type)


class Field(lib.types.GenericAlias, lib.t.Generic[AnyTypeCo]):
    """Single Generic alias type."""

    @lib.t.overload
    def __new__(cls, origin: type, args: type[AnyTypeCo]) -> lib.Self: ...
    @lib.t.overload
    def __new__(cls, origin: type, args: AnyTypeCo) -> lib.Self: ...
    def __new__(
        cls,
        origin: type,
        args: lib.t.Union[type[AnyTypeCo], AnyTypeCo]
        ) -> lib.Self:
        return super().__new__(cls, origin, args)

    def __repr__(self) -> str:
        ftypes = utl.check.expand_types(self)
        _delim = (
            ' | '
            if core.typ.utl.check.is_union(self)
            else ', '
            )
        _ftypes = _delim.join(
            (
                getattr(t, '__name__', 'Any')
                for t
                in ftypes
                )
            )
        return f'Field[{_ftypes}]'



"""Objects utility functions."""

__all__ = (
    'get_enumerations_from_fields',
    'get_fields_for_hash',
    'is_public_field',
    'is_valid_keyword',
    )

from . import cfg
from . import lib
from . import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""


@lib.functools.cache
def is_public_field(f: str) -> bool:
    """Return if field name is public."""

    return not (
        (f in Constants.FORBIDDEN_KEYWORDS)
        or f.startswith('_')
        )


@lib.functools.cache
def is_valid_keyword(f: str) -> bool:
    """Return if field name is allowed."""

    return f not in (
        set(Constants.FORBIDDEN_KEYWORDS)
        | set(Constants.BASE_ATTRS)
        )


def get_enumerations_from_fields(
    fields: typ.DataClassFields,
    ) -> dict[typ.string[typ.snake_case], tuple[typ.Primitive, ...]]:
    """
    Return dict containing all enums for object.

    ---

    Automatically appends `None` to any enums for an `Optional` type.

    """

    d: dict[typ.string[typ.snake_case], tuple[typ.Primitive, ...]] = {}
    for k, field in fields.items():
        if isinstance((enum_ := field.get('enum')), lib.enum.EnumMeta):
            d[k] = tuple([e.value for e in enum_._member_map_.values()])
        elif typ.utl.check.is_array(enum_):
            d[k] = tuple(enum_)
        if (
            k in d
            and isinstance(
                None,
                typ.utl.check.get_checkable_types(field.type_)
                )
            and None not in d[k]
            ):
            d[k] = (*d[k], None)

    return d


def get_fields_for_hash(
    __fields: typ.DataClassFields
    ) -> tuple[typ.string[typ.snake_case], ...]:
    """
    Filter to set of minimum fields required to compute a unique hash \
    for their owner object.

    ---

    Fields used must be of primitive types, for these purposes: \
    `bool | float | int | None | str`.

    Fields ending in the following will be used [in the following \
    order of precedence]:

    1. `'*id' | '*key'`
    2. `'*name'`

    For example, for an object with fields `'id_'` and \
    `'_common_name_'`, this function would return `('id_', )`, as \
    `'id_'` takes precedence over `'_common_name_'`.

    If no fields are named in ways that suggest they can be used to \
    determine the uniqueness of the object, all fields will instead be \
    returned.

    """

    id_fields: list[typ.string[typ.snake_case]] = []
    name_fields: list[typ.string[typ.snake_case]] = []
    primitive_fields: list[typ.string[typ.snake_case]] = []

    for f, field in __fields.items():
        if (
            isinstance(field.type_, (lib.t.ForwardRef, str))
            or not all(
                typ.utl.check.is_primitive(sub_tp)
                for sub_tp
                in typ.utl.check.get_checkable_types(field)
                )
            ):  # pragma: no cover
            continue
        elif (s := f.strip('_').lower()).endswith('id'):
            id_fields.append(f)
        elif s.endswith('key'):
            id_fields.append(f)
        elif s.startswith('name') or s.endswith('name'):
            name_fields.append(f)
        else:
            primitive_fields.append(f)

    if id_fields:
        fields_for_hash = tuple(id_fields)
    elif name_fields:
        fields_for_hash = tuple(name_fields)
    elif primitive_fields:
        fields_for_hash = tuple(primitive_fields)
    else:
        fields_for_hash = tuple(__fields)

    return fields_for_hash



"""Field module."""

__all__ = (
    'Field',
    )

from ... import core

from .. import cfg
from .. import enm
from .. import exc
from .. import lib
from .. import objs
from .. import typ

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from .. import queries


class Constants(cfg.Constants):
    """Constant values specific to this file."""

    FACTORY_CACHE: dict[str, lib.t.Callable[[], lib.t.Any]] = {}


class Field(objs.Object, lib.t.Generic[typ.AnyType]):
    """
    Simple field object.

    ---

    ## Querying

    Queries for `Objects` can be generated from their fields \
    using the following comparison operators:

    * `field_1_eq_filter = Object.field_1 == 'test_value_123'`
    * `field_1_ne_filter = Object.field_1 != 'test_value_123'`
    * `field_1_ge_filter = Object.field_1 >= 'test_value_123'`
    * `field_1_gt_filter = Object.field_1 > 'test_value_123'`
    * `field_1_le_filter = Object.field_1 <= 'test_value_123'`
    * `field_1_lt_filter = Object.field_1 < 'test_value_123'`

    And the following special operators:

    * `field_1_contains_filter = Object.field_1 << 'test_value_123'`
    * `field_1_similarity_filter = Object.field_1 % 'test_value_123'`
    * `field_1_similarity_filter_with_threshold = Object.field_1 % ('test_value_123', 0.8)`

    Queries may be chained together using the `&` and `|` bitwise \
    operators, corresponding to `and` and `or` clauses respectively.

    Additionally, the invert (`~`) operator may be prefixed to any \
    Query to match the opposite of any conditions specified \
    instead.

    Queries also support optional result limiting and sorting:

    * Result limits can be specified by setting the `limit` field.
    * Results can be sorted any number of times using the `+=` and `-=` \
    operators.

    ---

    ### Example

    ```py
    query: Query = (
        (
            (Object.integer_field >= 1)
            | (Object.string_field % ('test', 0.75))
            )
        & ~(Object.list_field << 'test')
        ) += 'string_field' -= 'integer_field'

    ```

    In the example above, the query would match any `Object` for which \
    the string `'test'` is `not` a member of `list_field` and for which \
    either the value for `integer_field` is greater than or equal to `1` \
    or the value for `string_field` is at least `75%` similar to `'test'`. \
    Results would then be sorted first in `ascending` order on `string_field`, \
    then in `descending` order on `integer_field`.

    ---

    ## Parameters

    Specify parameters to constrain values allowed for the field \
    and control its behavior.


    ```py
    name: str = None

    ```

    Field Name.
    Sourced from / overwritten by attribute name.


    ```py
    type: type[lib.t.Any] = None

    ```

    Type of value.
    Sourced from / overwritten by type annotation.


    ```py
    default: lib.t.Any = None

    ```

    Default value for field.
    Sourced from / overwritten by attribute value.
    MUST be an instance of field `type` or `None`.


    ```py
    required: bool = False

    ```

    Whether or not the field SHOULD be required.
    Default behavior changes to assume `True` if \
    no attribute value is specified for the field.


    ```py
    enum: deque | frozenset | list | tuple | set | Enum = None

    ```

    Sequence of which field value SHOULD be a member, unless \
    `"*"` is included in the sequence, in which case ANY value \
    MAY be allowed, in addition to those explicitly specified.
    `None` is assumed to be allowed if the field is nullable.


    ```py
    min_length: int = None

    ```

    Specify `len(value)` SHOULD be `>=` minimum.
    Field type MUST be `str` if specified.


    ```py
    max_length: int = None

    ```

    Specify `len(value)` SHOULD be `<=` maximum.
    Field type MUST be `str` if specified.


    ```py
    minimum: float = None

    ```

    Specify value SHOULD be `>=` minimum.
    Field type MUST be numeric if specified.


    ```py
    exclusive_minimum: bool = False

    ```

    Set `True` to specify value SHOULD be `>` minimum.
    Field minimum MUST also be specified.


    ```py
    maximum: float = None

    ```

    Specify value SHOULD be `<=` maximum.
    Field type MUST be numeric if specified.


    ```py
    exclusive_maximum: bool = False

    ```

    Set `True` to specify value SHOULD be `<` maximum.
    Field maximum MUST also be specified.


    ```py
    multiple_of: float = None

    ```

    Specify `value % multiple_of` SHOULD be `0`.
    Field type MUST be numeric if specified.


    ```py
    pattern: str = None

    ```

    Specify a Regex pattern for which the value SHOULD match.
    Field type MUST be `str` if specified.


    ```py
    min_items: int = None

    ```

    Specify `len(value)` SHOULD be `>=` min_items.
    Field type MUST be `deque | frozenset | list | tuple | set` if specified.


    ```py
    max_items: int = None

    ```

    Specify `len(value)` SHOULD be `<=` max_items.
    Field type MUST be `deque | frozenset | list | tuple | set` if specified.


    ```py
    unique_items: bool = False

    ```

    Specify all elements of value SHOULD be unique.
    Field type MUST be `deque | frozenset | list | tuple | set` if specified.


    ```py
    read_only: bool = False

    ```

    Specify this field SHOULD only be available to read \
    operations (like `GET` http calls).


    ```py
    write_only: bool = False

    ```

    Specify this field SHOULD only be available to write \
    operations (like `PATCH`, `POST`, or `PUT` http calls).

    """

    name: 'Field[str]' = None
    type_: 'Field[type[typ.AnyType]]' = None
    default: 'Field[typ.AnyType]' = None
    required: 'Field[bool]' = False
    enum: 'Field[typ.Enum]' = None
    min_length: 'Field[int]' = None
    max_length: 'Field[int]' = None
    minimum: 'Field[float]' = None
    exclusive_minimum: 'Field[bool]' = False
    maximum: 'Field[float]' = None
    exclusive_maximum: 'Field[bool]' = False
    multiple_of: 'Field[float]' = None
    pattern: 'Field[str]' = None
    min_items: 'Field[int]' = None
    max_items: 'Field[int]' = None
    unique_items: 'Field[bool]' = False
    read_only: 'Field[bool]' = False
    write_only: 'Field[bool]' = False

    @lib.t.overload
    def __get__(
        self,
        object_: None,
        dtype: type['objs.Object']
        ) -> 'Field[typ.AnyType]': ...
    @lib.t.overload
    def __get__(
        self,
        object_: 'objs.Object',
        dtype: type['objs.Object']
        ) -> typ.AnyType: ...
    def __get__(
        self,
        object_: lib.t.Optional['objs.Object'],
        dtype: type['objs.Object']
        ) -> 'Field[typ.AnyType]' | typ.AnyType:  # pragma: no cover
        return self

    def __set__(
        self,
        __object: lib.t.Any,
        __value: typ.AnyType
        ) -> lib.t.Optional[lib.Never]:
        if isinstance(__value, str):
            object.__setattr__(__object, self.name, self.parse(__value))
        elif isinstance(
            __value,
            typ.utl.check.get_checkable_types(self.type_)
            ):
            object.__setattr__(__object, self.name, __value)
        else:
            raise exc.IncorrectTypeError(self.name, self.type_, __value)
        return None

    @lib.t.overload
    def __init__(
        self,
        class_as_dict: lib.t.Optional[dict[typ.string[typ.snake_case], lib.t.Any]] = None,
        /,
        *,
        type_: type[typ.Type] = None,
        default: typ.Type = None,
        required: bool = False,
        enum: typ.Enum = None,
        min_length: int = None,
        max_length: int = None,
        minimum: float = None,
        exclusive_minimum: bool = False,
        maximum: float = None,
        exclusive_maximum: bool = False,
        multiple_of: float = None,
        pattern: str = None,
        min_items: int = None,
        max_items: int = None,
        unique_items: bool = False,
        read_only: bool = False,
        write_only: bool = False,
        **kwargs: lib.t.Any
        ): ...
    @lib.t.overload
    def __init__(
        self,
        class_as_dict: lib.t.Optional[dict[typ.string[typ.snake_case], lib.t.Any]] = None,
        /,
        *,
        type_: type[typ.AnyType] = None,
        default: typ.AnyType = None,
        required: bool = False,
        enum: typ.Enum = None,
        min_length: int = None,
        max_length: int = None,
        minimum: float = None,
        exclusive_minimum: bool = False,
        maximum: float = None,
        exclusive_maximum: bool = False,
        multiple_of: float = None,
        pattern: str = None,
        min_items: int = None,
        max_items: int = None,
        unique_items: bool = False,
        read_only: bool = False,
        write_only: bool = False,
        **kwargs: lib.t.Any
        ): ...
    def __init__(
        self,
        class_as_dict: lib.t.Optional[dict[typ.string[typ.snake_case], lib.t.Any]] = None,
        /,
        *,
        type_: type | type[typ.Type] | type[typ.AnyType] = None,
        default: lib.t.Any | typ.Type | typ.AnyType = None,
        required: bool = False,
        enum: typ.Enum = None,
        min_length: int = None,
        max_length: int = None,
        minimum: float = None,
        exclusive_minimum: bool = False,
        maximum: float = None,
        exclusive_maximum: bool = False,
        multiple_of: float = None,
        pattern: str = None,
        min_items: int = None,
        max_items: int = None,
        unique_items: bool = False,
        read_only: bool = False,
        write_only: bool = False,
        **kwargs: lib.t.Any
        ):
        if class_as_dict is not None:
            kwargs |= class_as_dict  # type: ignore[arg-type]
        else:
            kwargs |= dict(
                type_=lib.t.cast(
                    type[typ.AnyType],
                    kwargs.pop('type_', kwargs.pop('type', type_))
                    ),
                default=default,
                required=required,
                enum=enum,
                min_length=min_length,
                max_length=max_length,
                minimum=minimum,
                exclusive_minimum=exclusive_minimum,
                maximum=maximum,
                exclusive_maximum=exclusive_maximum,
                multiple_of=multiple_of,
                pattern=pattern,
                min_items=min_items,
                max_items=max_items,
                unique_items=unique_items,
                read_only=read_only,
                write_only=write_only,
                )

        ckwargs = {
            cname: value
            for name, value
            in kwargs.items()
            if (cname := core.strings.utl.cname_for(name, self.fields))
            }

        if isinstance(class_as_dict, lib.t.Mapping):
            class_as_cdict = {
                cname: value
                for name, value
                in class_as_dict.items()
                if (cname := core.strings.utl.cname_for(name, self.fields))
                }
            ckwargs |= class_as_cdict

        fname: typ.string[typ.snake_case]
        for fname, field in self.__dataclass_fields__.items():
            if fname not in ckwargs:
                ckwargs[fname] = field['default']

        for cname_, value in ckwargs.items():
            setattr(self, cname_, value)

        self.__post_init__()

    def __field_hash__(self) -> int:
        return hash(
            ''.join(
                (
                    self.__class__.__name__,
                    repr(self.type_) or str(),
                    repr(self.default) or str(),
                    self.name or str()
                    )
                )
            )

    @lib.t.overload
    def __eq__(
        self,
        other: 'typ.AnyField[lib.t.Any]'
        ) -> bool: ...
    @lib.t.overload
    def __eq__(self: object, other: object) -> bool: ...
    @lib.t.overload
    def __eq__(
        self,
        other: lib.t.Any
        ) -> 'bool | queries.EqQueryCondition | lib.Never': ...
    def __eq__(
        self,
        other: 'object | lib.t.Any | typ.AnyField[lib.t.Any]'
        ) -> 'bool | queries.EqQueryCondition | lib.Never':
        if typ.utl.check.is_field(other):
            return self.__field_hash__() == other.__field_hash__()
        self._validate_comparison(other)
        from .. import queries
        q: queries.EqQueryCondition = (
            queries.EqQueryCondition(
                field=self.name.rstrip('_'),
                eq=other
                )
            )
        return q

    @lib.t.overload
    def __ne__(
        self,
        other: 'typ.AnyField[lib.t.Any]'
        ) -> bool: ...
    @lib.t.overload
    def __ne__(self: object, other: object) -> bool: ...
    @lib.t.overload
    def __ne__(
        self,
        other: lib.t.Any
        ) -> 'bool | queries.NeQueryCondition | lib.Never': ...
    def __ne__(
        self,
        other: 'object | lib.t.Any | typ.AnyField[lib.t.Any]'
        ) -> 'bool | queries.NeQueryCondition | lib.Never':
        if typ.utl.check.is_field(other):
            return self.__field_hash__() != other.__field_hash__()
        self._validate_comparison(other)
        from .. import queries
        q: queries.NeQueryCondition = (
            queries.NeQueryCondition(
                field=self.name.rstrip('_'),
                ne=other
                )
            )
        return q

    @lib.t.overload
    def __mod__(
        self,
        params: tuple[typ.AnyType, float]
        ) -> 'queries.SimilarQueryCondition': ...
    @lib.t.overload
    def __mod__(
        self,
        params: typ.AnyType
        ) -> 'queries.SimilarQueryCondition': ...
    @lib.t.overload
    def __mod__(
        self,
        params: tuple[lib.t.Any, ...]
        ) -> 'queries.SimilarQueryCondition | lib.Never': ...
    @lib.t.overload
    def __mod__(
        self,
        params: lib.t.Any
        ) -> 'queries.SimilarQueryCondition | lib.Never': ...
    def __mod__(
        self,
        params: tuple[typ.AnyType, float] | tuple[lib.t.Any, ...] | lib.t.Any
        ) -> 'queries.SimilarQueryCondition | lib.Never':
        if isinstance(params, tuple):
            value, threshold = params
            if not isinstance(threshold, float):
                threshold = enm.MatchThreshold.default.value
        else:
            value = params
            threshold = enm.MatchThreshold.default.value
        self._validate_iterable_comparison(value)
        self._validate_comparison(value)
        from .. import queries
        q: queries.SimilarQueryCondition = (
            queries.SimilarQueryCondition(
                field=self.name.rstrip('_'),
                like=value,
                threshold=threshold
                )
            )
        return q

    @lib.t.overload
    def __lshift__(
        self,
        value: typ.obj.ObjectLike
        ) -> lib.Self | lib.Never: ...
    @lib.t.overload
    def __lshift__(
        self,
        value: lib.t.Any
        ) -> lib.t.Union[
            'queries.ContainsQueryCondition',
            'Field[lib.t.Any]',
            lib.Never
            ]: ...
    def __lshift__(
        self,
        value: 'lib.t.Any | typ.obj.ObjectLike'
        ) -> lib.t.Union[
            'queries.ContainsQueryCondition',
            'Field[lib.t.Any]',
            lib.Never
            ]:
        if typ.utl.check.is_field(value):
            return super().__lshift__(value)
        self._validate_iterable_comparison(value)
        from .. import queries
        q: queries.ContainsQueryCondition = (
            queries.ContainsQueryCondition(
                field=self.name.rstrip('_'),
                contains=value
                )
            )
        return q

    @lib.t.overload
    def __gt__(
        self,
        value: typ.AnyType
        ) -> 'queries.GtQueryCondition': ...
    @lib.t.overload
    def __gt__(
        self,
        value: lib.t.Any
        ) -> 'queries.GtQueryCondition | lib.Never': ...
    def __gt__(
        self,
        value: typ.AnyType | lib.t.Any
        ) -> 'queries.GtQueryCondition | lib.Never':
        self._validate_comparison(value)
        from .. import queries
        q: queries.GtQueryCondition = (
            queries.GtQueryCondition(
                field=self.name.rstrip('_'),
                gt=value
                )
            )
        return q

    @lib.t.overload
    def __ge__(
        self,
        value: typ.AnyType
        ) -> 'queries.GeQueryCondition': ...
    @lib.t.overload
    def __ge__(
        self,
        value: lib.t.Any
        ) -> 'queries.GeQueryCondition | lib.Never': ...
    def __ge__(
        self,
        value: typ.AnyType | lib.t.Any
        ) -> 'queries.GeQueryCondition | lib.Never':
        self._validate_comparison(value)
        from .. import queries
        q: queries.GeQueryCondition = (
            queries.GeQueryCondition(
                field=self.name.rstrip('_'),
                ge=value
                )
            )
        return q

    @lib.t.overload
    def __lt__(
        self,
        value: typ.AnyType
        ) -> 'queries.LtQueryCondition': ...
    @lib.t.overload
    def __lt__(
        self,
        value: lib.t.Any
        ) -> 'queries.LtQueryCondition | lib.Never': ...
    def __lt__(
        self,
        value: typ.AnyType | lib.t.Any
        ) -> 'queries.LtQueryCondition | lib.Never':
        self._validate_comparison(value)
        from .. import queries
        q: queries.LtQueryCondition = (
            queries.LtQueryCondition(
                field=self.name.rstrip('_'),
                lt=value
                )
            )
        return q

    @lib.t.overload
    def __le__(
        self,
        value: typ.AnyType
        ) -> 'queries.LeQueryCondition': ...
    @lib.t.overload
    def __le__(
        self,
        value: lib.t.Any
        ) -> 'queries.LeQueryCondition | lib.Never': ...
    def __le__(
        self,
        value: typ.AnyType | lib.t.Any
        ) -> 'queries.LeQueryCondition | lib.Never':
        self._validate_comparison(value)
        from .. import queries
        q: queries.LeQueryCondition = (
            queries.LeQueryCondition(
                field=self.name.rstrip('_'),
                le=value
                )
            )
        return q

    @lib.t.overload
    def _validate_comparison(
        self,
        value: lib.t.Optional[typ.AnyType]
        ) -> None: ...
    @lib.t.overload
    def _validate_comparison(
        self,
        value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]: ...
    def _validate_comparison(
        self,
        value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]:
        if (
            value is not None
            and not isinstance(
                value,
                typ.utl.check.get_checkable_types(self.type_)
                )
            ):
            raise exc.InvalidComparisonTypeError(
                self.name,
                self.type_,
                value
                )
        else:
            return None

    def _validate_iterable_comparison(
        self,
        value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]:
        if not any(
            issubclass(tp, lib.t.Iterable)
            for tp
            in typ.utl.check.get_checkable_types(self.type_)
            ):
            raise exc.InvalidContainerComparisonTypeError(
                self.name,
                self.type_,
                value
                )
        else:
            return None

    def parse(
        self,
        value: lib.t.Any,
        raise_validation_error: bool = True,
        ) -> lib.t.Optional[typ.AnyType] | lib.Never:
        """
        Return correctly typed value if possible, `None` otherwise, or \
        [optionally] raise an error if an invalid value is passed, the \
        method's default behavior.

        """

        parsed = core.codecs.utl.parse(value, self.type_)
        if isinstance(parsed, core.codecs.enm.ParseErrorRef):
            if raise_validation_error:
                raise exc.TypeValidationError(self.name, self.type_, parsed)
            else:
                return None
        else:
            return parsed

    @property
    def factory(self) -> lib.t.Callable[[], typ.AnyType]:
        """Return callable returning default value for field."""

        if (
            (
                key := Constants.DELIM.join(
                    (
                        str(self.name).lower(),
                        getattr(
                            self.type_,
                            '__name__',
                            self.type_.__class__.__name__
                            ),
                        str(self.default),
                        'factory'
                        )
                    )
                )
            not in Constants.FACTORY_CACHE
            ):
            if (
                typ.utl.check.is_immutable_type(
                    self.type_ or type(self.default)
                    )
                ):
                Constants.FACTORY_CACHE[key] = lambda: self.default
            elif callable(self.default):
                Constants.FACTORY_CACHE[key] = lambda: self.default()  # type: ignore[operator]
            else:
                Constants.FACTORY_CACHE[key] = lambda: lib.copy.deepcopy(self.default)

        value: lib.t.Callable[[], typ.AnyType] = Constants.FACTORY_CACHE[key]
        return value



"""Objects enumerations."""

from .. import core

__all__ = (
    'MatchThreshold',
    'SortDirection',
    *core.enm.__all__,
    )

from .. core . enm import *

from . import lib


class MatchThreshold(lib.enum.Enum):
    """Minimum Query Similarity Enumeration."""

    default = 0.85


class SortDirection(lib.enum.Enum):
    """Sort Direction Enumeration."""

    asc  = 'asc'
    desc = 'desc'



"""Strings objects."""

__all__ = (
    'Pattern',
    'KeyValueRedactionPattern',
    'KeyValueRedactionPatterns',
    'RedactionPattern',
    'RedactionPatterns',
    'StringWrapper',
    )

from . import cfg
from . import lib


class Constants(cfg.Constants):
    """Constants specific to this file."""


class Pattern:
    """Compiled regex patterns."""

    SnakeToCamelReplacements = lib.re.compile(r'(_[a-z0-9])')
    """
    Matches all lower case alphanumeric characters following any \
    non-leading underscore.

    ---

    Note: match is inclusive of underscores to improve substitution \
    performance.

    """

    CamelToSnakeReplacements = lib.re.compile(
        r'[A-Z0-9]([0-9]+|[a-z]+|([0-9][a-z])+)'
        )
    """Matches all Title Case and numeric components."""

    camelCase = lib.re.compile(r'^[a-z]+((\d)|([A-Z0-9][a-z0-9]{1,128})){0,32}$')
    """
    Matches strict [lower] camelCase (i.e. RESTful casing) according to \
    the [Google Java Style Guide](https://google.github.io/styleguide/javaguide.html#s5.3-camel-case).

    ---

    Unlike Google, does NOT allow for an optional uppercase character at \
    the end of the string.

    Strings with more than32IndividualWords or \
    withWordsLongerThan128Characters will not be matched.

    """

    snake_case = lib.re.compile(r'^[a-z0-9_]{1,4096}$')
    """
    Matches strict [lower] snake_case (i.e. python instance \
    / attribute / function casing).

    ---

    Strings longer than 4096 characters will not be matched.

    """

    DateTime = lib.re.compile(
        r'[0-9]{4}-[0-9]{2}-[0-9]{2}'
        '('
            r'[ T][0-9]{2}:[0-9]{2}:[0-9]{2}'
            r'(\.([0-9]{1,6}))?'
        ')?'
        r'([+-][0-9]{2}:[0-9]{2})?'
        )
    """
    Matches valid python `datetime` strings.

    ---

    Note: validity is determined by parsability `fromisoformat()`.

    """

    Number = lib.re.compile(
        '^'
        '('
            r'[+-]?'
            r'([0-9](_?[0-9]){0,63})?'
            r'(\.)?'
            r'[0-9](_?[0-9]){0,63}'
            r'(e[+-]?[0-9](_?[0-9]){0,63})?'
        ')'
        '('
            'j'
            '|'
            '('
                r'[+-]'
                r'([0-9](_?[0-9]){0,63})?'
                r'(\.)?'
                r'[0-9](_?[0-9]){0,63}'
                r'(e[+-]?[0-9](_?[0-9]){0,63})?'
                'j'
            ')'
        ')?'
        '$'
        )
    """
    Matches integers, floats, scientific notation, and complex numbers.

    ---

    Supports precision up to 64 digits either side of a decimal point.

    Recognizes valid, pythonic underscore usage as well.

    """


class RedactionPattern:
    """
    Regex patterns to redact common, sensitive data.

    Sourced from trivy / aqua.
    https://github.com/aquasecurity/trivy/blob/main/pkg/fanal/secret/builtin-rules.go

    """

    aws_access_key_id = lib.re.compile(
        r'(?P<secret>(A3T[A-Z0-9]|AKIA|AGPA|AIDA|AROA|'
        r'AIPA|ANPA|ANVA|ASIA)[A-Z0-9]{16})'
        )
    aws_secret_access_key = lib.re.compile(
        r'(?i)(^|\s+)(?P<secret>[A-Za-z0-9\/\+=]{40})'
        )
    github_pat = lib.re.compile(r'ghp_[0-9a-zA-Z]{36}')
    github_oauth = lib.re.compile(r'gho_[0-9a-zA-Z]{36}')
    github_app_token = lib.re.compile(r'(ghu|ghs)_[0-9a-zA-Z]{36}')
    github_refresh_token = lib.re.compile(r'ghr_[0-9a-zA-Z]{76}')
    github_fine_grained_pat = lib.re.compile(
        r'github_pat_[a-zA-Z0-9]{22}_[a-zA-Z0-9]{59}'
        )
    gitlab_pat = lib.re.compile(r'glpat-[0-9a-zA-Z\-\_]{20}')
    private_key = lib.re.compile(
        r'(?i)-----\s*?BEGIN[ A-Z0-9_-]*?PRIVATE KEY( BLOCK)?\s*?-----[\s]*?'
        r'(?P<secret>[\sA-Za-z0-9=+/\\\r\n]+)[\s]*?-----\s*?END[ A-Z0-9_-]*?'
        r' PRIVATE KEY( BLOCK)?\s*?-----'
        )
    shopify_token = lib.re.compile(r'shp(ss|at|ca|pa)_[a-fA-F0-9]{32}')
    slack_access_token = lib.re.compile(r'xox[baprs]-([0-9a-zA-Z]{10,48})')
    stripe_publishable_token = lib.re.compile(
        r'(?i)pk_(test|live)_[0-9a-z]{10,32}'
        )
    stripe_secret_token = lib.re.compile(
        r'(?i)sk_(test|live)_[0-9a-z]{10,32}'
        )
    pypi_upload_token = lib.re.compile(
        r'pypi-AgEIcHlwaS5vcmc[A-Za-z0-9\-_]{50,1000}'
        )
    gcp_service_account = lib.re.compile(r'\"type\": \"service_account\"')
    heroku_api_key = lib.re.compile(
        r'(?i)(?P<key>heroku[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]'
        r'{4}-[0-9A-F]{12})[\'\"]'
        )
    slack_web_hook = lib.re.compile(
        r'https:\/\/hooks\.slack\.com\/services\/[A-Za-z0-9+\/]{44,48}'
        )
    twilio_api_key = lib.re.compile(r'SK[0-9a-fA-F]{32}')
    age_secret_key = lib.re.compile(
        r'AGE-SECRET-KEY-1[QPZRY9X8GF2TVDW0S3JN54KHCE6MUA7L]{58}'
        )
    facebook_token = lib.re.compile(
        r'(?i)(?P<key>facebook[a-z0-9_ .\-,]{0,25})'
        r'(=|>|:=|\|\|:|<=|=>|:).{0,5}[\'\"](?P<secret>[a-f0-9]{32})[\'\"]'
        )
    twitter_token = lib.re.compile(
        r'(?i)(?P<key>twitter[a-z0-9_ .\-,]{0,25})'
        r'(=|>|:=|\|\|:|<=|=>|:).{0,5}[\'\"](?P<secret>[a-f0-9]{35,44})[\'\"]'
        )
    adobe_client_id = lib.re.compile(
        r'(?i)(?P<key>adobe[a-z0-9_ .\-,]{0,25})'
        r'(=|>|:=|\|\|:|<=|=>|:).{0,5}[\'\"](?P<secret>[a-f0-9]{32})[\'\"]'
        )
    adobe_client_secret = lib.re.compile(r'(?i)(p8e-)[a-z0-9]{32}')
    alibaba_access_key_id = lib.re.compile(
        r'(?i)([^0-9A-Za-z]|^)(?P<secret>(LTAI)[a-z0-9]{20})([^0-9A-Za-z]|$)'
        )
    alibaba_secret_key = lib.re.compile(
        r'(?i)(?P<key>alibaba[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{30})[\'\"]'
        )
    asana_client_id = lib.re.compile(
        r'(?i)(?P<key>asana[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[0-9]{16})[\'\"]'
        )
    asana_client_secret = lib.re.compile(
        r'(?i)(?P<key>asana[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{32})[\'\"]'
        )
    atlassian_api_token = lib.re.compile(
        r'(?i)(?P<key>atlassian[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{24})[\'\"]'
        )
    bitbucket_client_id = lib.re.compile(
        r'(?i)(?P<key>bitbucket[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{32})[\'\"]'
        )
    bitbucket_client_secret = lib.re.compile(
        r'(?i)(?P<key>bitbucket[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9_\-]{64})[\'\"]'
        )
    beamer_api_token = lib.re.compile(
        r'(?i)(?P<key>beamer[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>b_[a-z0-9=_\-]{44})[\'\"]'
        )
    clojars_api_token = lib.re.compile(r'(?i)(CLOJARS_)[a-z0-9]{60}')
    contentful_delivery_api_token = lib.re.compile(
        r'(?i)(?P<key>contentful[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9\-=_]{43})[\'\"]'
        )
    databricks_api_token = lib.re.compile(r'dapi[a-h0-9]{32}')
    discord_api_token = lib.re.compile(
        r'(?i)(?P<key>discord[a-z0-9_ .\-,]{0,25})'
        r'(=|>|:=|\|\|:|<=|=>|:).{0,5}[\'\"](?P<secret>[a-h0-9]{64})[\'\"]'
        )
    discord_client_id = lib.re.compile(
        r'(?i)(?P<key>discord[a-z0-9_ .\-,]{0,25})'
        r'(=|>|:=|\|\|:|<=|=>|:).{0,5}[\'\"](?P<secret>[0-9]{18})[\'\"]'
        )
    discord_client_secret = lib.re.compile(
        r'(?i)(?P<key>discord[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9=_\-]{32})[\'\"]'
        )
    doppler_api_token = lib.re.compile(
        r'(?i)[\'\"](dp\.pt\.)[a-z0-9]{43}[\'\"]'
        )
    dropbox_api_secret = lib.re.compile(
        r'(?i)(dropbox[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"]([a-z0-9]{15})[\'\"]'
        )
    dropbox_short_lived_api_token = lib.re.compile(
        r'(?i)(dropbox[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](sl\.[a-z0-9\-=_]{135})[\'\"]'
        )
    dropbox_long_lived_api_token = lib.re.compile(
        r'(?i)(dropbox[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"][a-z0-9]{11}(AAAAAAAAAA)[a-z0-9\-_=]{43}[\'\"]'
        )
    duffel_api_token = lib.re.compile(
        r'(?i)[\'\"]duffel_(test|live)_[a-z0-9_-]{43}[\'\"]'
        )
    dynatrace_api_token = lib.re.compile(
        r'(?i)[\'\"]dt0c01\.[a-z0-9]{24}\.[a-z0-9]{64}[\'\"]'
        )
    easypost_api_token = lib.re.compile(
        r'(?i)[\'\"]EZ[AT]K[a-z0-9]{54}[\'\"]'
        )
    fastly_api_token = lib.re.compile(
        r'(?i)(?P<key>fastly[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9\-=_]{32})[\'\"]'
        )
    finicity_client_secret = lib.re.compile(
        r'(?i)(?P<key>finicity[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{20})[\'\"]'
        )
    finicity_api_token = lib.re.compile(
        r'(?i)(?P<key>finicity[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-f0-9]{32})[\'\"]'
        )
    flutterwave_public_key = lib.re.compile(
        r'(?i)FLW(PUB|SEC)K_TEST-[a-h0-9]{32}-X'
        )
    flutterwave_enc_key = lib.re.compile(r'FLWSECK_TEST[a-h0-9]{12}')
    frameio_api_token = lib.re.compile(r'(?i)fio-u-[a-z0-9\-_=]{64}')
    gocardless_api_token = lib.re.compile(
        r'(?i)[\'\"]live_[a-z0-9\-_=]{40}[\'\"]'
        )
    grafana_api_token = lib.re.compile(
        r'(?i)[\'\"]eyJrIjoi[a-z0-9\-_=]{72,92}[\'\"]'
        )
    hashicorp_tf_api_token = lib.re.compile(
        r'(?i)[\'\"][a-z0-9]{14}\.atlasv1\.[a-z0-9\-_=]{60,70}[\'\"]'
        )
    hubspot_api_token = lib.re.compile(
        r'(?i)(?P<key>hubspot[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-h0-9]{8}-[a-h0-9]{4}-[a-h0-9]{4}-'
        r'[a-h0-9]{4}-[a-h0-9]{12})[\'\"]'
        )
    intercom_api_token = lib.re.compile(
        r'(?i)(?P<key>intercom[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9=_]{60})[\'\"]'
        )
    intercom_client_secret = lib.re.compile(
        r'(?i)(?P<key>intercom[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-h0-9]{8}-[a-h0-9]{4}-[a-h0-9]{4}-'
        r'[a-h0-9]{4}-[a-h0-9]{12})[\'\"]'
        )
    ionic_api_token = lib.re.compile(
        r'(?i)(ionic[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}'
        r'[\'\"](ion_[a-z0-9]{42})[\'\"]'
        )
    jwt_token = lib.re.compile(
        r'ey[a-zA-Z0-9]{17,}\.ey[a-zA-Z0-9\/\\_-]{17,}\.'
        r'(?:[a-zA-Z0-9\/\\_-]{10,}={0,2})?'
        )
    linear_api_token = lib.re.compile(
        r'(?i)lin_api_[a-z0-9]{40}'
        )
    linear_client_secret = lib.re.compile(
        r'(?i)(?P<key>linear[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).'
        r'{0,5}[\'\"](?P<secret>[a-f0-9]{32})[\'\"]'
        )
    lob_api_key = lib.re.compile(
        r'(?i)(?P<key>lob[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).'
        r'{0,5}[\'\"](?P<secret>(live|test)_[a-f0-9]{35})[\'\"]'
        )
    lob_pub_api_key = lib.re.compile(
        r'(?i)(?P<key>lob[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).'
        r'{0,5}[\'\"](?P<secret>(test|live)_pub_[a-f0-9]{31})[\'\"]'
        )
    mailchimp_api_key = lib.re.compile(
        r'(?i)(?P<key>mailchimp[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).'
        r'{0,5}[\'\"](?P<secret>[a-f0-9]{32}-us20)[\'\"]'
        )
    mailgun_token = lib.re.compile(
        r'(?i)(?P<key>mailgun[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).'
        r'{0,5}[\'\"](?P<secret>(pub)?key-[a-f0-9]{32})[\'\"]'
        )
    mailgun_signing_key = lib.re.compile(
        r'(?i)(?P<key>mailgun[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).'
        r'{0,5}[\'\"](?P<secret>[a-h0-9]{32}-[a-h0-9]{8}-[a-h0-9]{8})[\'\"]'
        )
    mapbox_api_token = lib.re.compile(
        r'(?i)(pk\.[a-z0-9]{60}\.[a-z0-9]{22})'
        )
    messagebird_api_token = lib.re.compile(
        r'(?i)(?P<key>messagebird[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{25})[\'\"]'
        )
    messagebird_client_id = lib.re.compile(
        r'(?i)(?P<key>messagebird[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-h0-9]{8}-[a-h0-9]{4}-[a-h0-9]{4}-[a-h0-9]'
        r'{4}-[a-h0-9]{12})[\'\"]'
        )
    new_relic_user_api_key = lib.re.compile(
        r'[\'\"](NRAK-[A-Z0-9]{27})[\'\"]'
        )
    new_relic_user_api_id = lib.re.compile(
        r'(?i)(?P<key>newrelic[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[A-Z0-9]{64})[\'\"]'
        )
    new_relic_browser_api_token = lib.re.compile(
        r'[\'\"](NRJS-[a-f0-9]{19})[\'\"]'
        )
    npm_access_token = lib.re.compile(r'(?i)[\'\"](npm_[a-z0-9]{36})[\'\"]')
    planetscale_password = lib.re.compile(r'(?i)pscale_pw_[a-z0-9\-_\.]{43}')
    planetscale_api_token = lib.re.compile(
        r'(?i)pscale_tkn_[a-z0-9\-_\.]{43}'
        )
    postman_api_token = lib.re.compile(r'(?i)PMAK-[a-f0-9]{24}\-[a-f0-9]{34}')
    pulumi_api_token = lib.re.compile(r'pul-[a-f0-9]{40}')
    rubygems_api_token = lib.re.compile(r'rubygems_[a-f0-9]{48}')
    sendgrid_api_token = lib.re.compile(r'(?i)SG\.[a-z0-9_\-\.]{66}')
    sendinblue_api_token = lib.re.compile(
        r'(?i)xkeysib-[a-f0-9]{64}\-[a-z0-9]{16}'
        )
    shippo_api_token = lib.re.compile(r'shippo_(live|test)_[a-f0-9]{40}')
    linkedin_client_secret = lib.re.compile(
        r'(?i)(?P<key>linkedin[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z]{16})[\'\"]'
        )
    linkedin_client_id = lib.re.compile(
        r'(?i)(?P<key>linkedin[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{14})[\'\"]'
        )
    twitch_api_token = lib.re.compile(
        r'(?i)(?P<key>twitch[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}[\'\"](?P<secret>[a-z0-9]{30})[\'\"]'
        )
    typeform_api_token = lib.re.compile(
        r'(?i)(?P<key>typeform[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:)'
        r'.{0,5}(?P<secret>tfp_[a-z0-9\-_\.=]{59})'
        )
    dockerconfig_secret = lib.re.compile(
        r'(?i)(\.(dockerconfigjson|dockercfg):\s*\|*\s*('
        r'?P<secret>(ey|ew)+[A-Za-z0-9\/\+=]+))'
        )


RedactionPatterns: dict[str, lib.re.Pattern[str]] = {
    k.upper(): v
    for k, v
    in RedactionPattern.__dict__.items()
    if isinstance(v, lib.re.Pattern)
    }
"""
Regex patterns to redact common, sensitive data.

Sourced from trivy / aqua.
https://github.com/aquasecurity/trivy/blob/main/pkg/fanal/secret/builtin-rules.go

"""


class KeyValueRedactionPattern:
    """Regex patterns to redact common, sensitive data."""

    api_key_token = lib.re.compile(
        r'(api|secret)+.?(key|token)',
        flags=lib.re.IGNORECASE
        )
    """Matches common names for api keys, tokens, etc."""

    authorization_header = lib.re.compile(
        r'(authorization|bearer)+',
        flags=lib.re.IGNORECASE
        )
    """Matches an authorization / bearer token header."""

    conn_string_password = lib.re.compile(r'(:\/\/)+\w+(:[^:@]+)@')
    """Matches passwords in database connection strings."""

    credit_card = lib.re.compile(
        r'\b(?:4[0-9]{12}(?:[0-9]{3})?|5[1-5][0-9]{14}|3[47][0-9]{13}|3'
        r'(?:0[0-5]|[68][0-9])[0-9]{11}|6(?:011|5[0-9]{2})[0-9]{12}|'
        r'(?:2131|1800|35\d{3})\d{11})\b'
        )
    """Matches an [ostensibly] valid credit card number."""


KeyValueRedactionPatterns: dict[str, lib.re.Pattern[str]] = {
    k.upper(): v
    for k, v
    in KeyValueRedactionPattern.__dict__.items()
    if isinstance(v, lib.re.Pattern)
    }
"""Regex patterns to redact common, sensitive data."""


StringWrapper = lib.textwrap.TextWrapper(
    width=Constants.WRAP_WIDTH,
    break_long_words=True,
    break_on_hyphens=True,
    max_lines=Constants.CUTOFF_LEN,
    expand_tabs=False,
    replace_whitespace=True,
    drop_whitespace=False,
    )
"""Wraps long strings."""



"""Codecs utility functions."""

__all__ = (
    'encode',
    'parse',
    'serialize',
    'try_decode',
    'try_parse_json',
    )

from .. import strings

from . import cfg
from . import enm
from . import lib
from . import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""


def serialize(value: lib.t.Any) -> str:
    """Convert value to string."""

    return lib.json.dumps(
        value,
        indent=Constants.INDENT,
        sort_keys=True,
        default=strings.utl.convert_for_repr
        )


def encode(value: lib.t.Any) -> typ.Serial:
    """
    JSON encode `value` using a corresponding encoder, otherwise returns \
    `repr(value)`."""

    if (encoder := Constants.ENCODERS.get(value.__class__)) is not None:
        return encoder(value)

    for _base in reversed(value.__class__.__bases__):
        for __base in reversed(_base.__mro__):
            if (encoder := Constants.ENCODERS.get(__base)) is not None:
                return encoder(value)

    return repr(value)


def try_decode(
    value: lib.t.Any,
    tp: type[typ.AnyType],
    ) -> typ.AnyType | enm.ParseErrorRef:
    """
    Attempt to parse value, returning `enm.ParseErrorRef` if error.

    """

    try:
        if typ.utl.check.is_literal(tp):
            literals = typ.utl.check.get_args(tp)
            literal: typ.AnyType
            if value in literals:
                literal = value
                return literal
            literal_tps = (type(literal) for literal in literals)
            for literal_tp in literal_tps:
                literal = try_decode(value, literal_tp)
                if literal in literals:
                    return literal
            return enm.ParseErrorRef.literal_decode
        elif isinstance(value, typ.utl.check.get_checkable_types(tp)):
            tp_value: typ.AnyType = value
            return tp_value
        elif isinstance(value, str):
            if typ.utl.check.is_bool_type(tp):
                if value.lower() in enm.Boolean._member_names_:
                    boolean: typ.AnyType = (
                        value.lower() == enm.Boolean.true.name
                        )
                    return boolean
                else:
                    return enm.ParseErrorRef.bool_decode
            elif typ.utl.check.is_number_type(tp):
                if strings.utl.is_valid_number_str(value):
                    num_tp: type[typ.AnyType] = tp
                    return num_tp(value)  # type: ignore[call-arg]
                else:
                    return enm.ParseErrorRef.number_decode
            elif (
                (is_datetime_tp := typ.utl.check.is_datetime_type(tp))
                or typ.utl.check.is_date_type(tp)
                ):
                if strings.utl.is_valid_datetime_str(value):
                    dt = (
                        lib.datetime.datetime.fromisoformat(value)
                        .replace(tzinfo=lib.datetime.timezone.utc)
                        )
                    if is_datetime_tp:
                        dt_value: typ.AnyType = dt
                        return dt_value
                    else:
                        date_value: typ.AnyType = dt.date()
                        return date_value
                else:
                    return enm.ParseErrorRef.datetime_decode
            elif typ.utl.check.is_none_type(tp):
                if value.lower() in enm.NoneAlias._member_names_:
                    none: typ.AnyType = None
                    return none
                else:
                    return enm.ParseErrorRef.null_decode
            else:  # pragma: no cover
                return tp(value)  # type: ignore[call-arg]
        else:
            return tp(value)  # type: ignore[call-arg]
    except:  # noqa: E722
        return enm.ParseErrorRef.value_decode


def try_parse_json(
    json_string: str
    ) -> typ.Serial | enm.ParseErrorRef:
    """
    Attempt to parse valid JSON string, returning \
    `enm.ParseErrorRef` if error.

    """

    try:
        deserialized: typ.Serial = lib.json.loads(json_string)
        return deserialized
    except:  # noqa: E722
        return enm.ParseErrorRef.invalid_json


@lib.t.overload
def parse(
    value: lib.t.Any,
    tp: type[typ.VariadicArrayType],
    ) -> typ.VariadicArrayType | enm.ParseErrorRef: ...
@lib.t.overload
def parse(
    value: lib.t.Any,
    tp: type[typ.ArrayType],
    ) -> typ.ArrayType | enm.ParseErrorRef: ...
@lib.t.overload
def parse(
    value: lib.t.Any,
    tp: type[typ.MappingType],
    ) -> typ.MappingType | enm.ParseErrorRef: ...
@lib.t.overload
def parse(
    value: lib.t.Any,
    tp: type[typ.Typed],
    ) -> typ.Typed | enm.ParseErrorRef: ...
@lib.t.overload
def parse(
    value: lib.t.Any,
    tp: type[typ.AnyType],
    ) -> typ.AnyType | enm.ParseErrorRef: ...
def parse(
    value: typ.AnyType,
    tp: (
        type[typ.VariadicArrayType]
        | type[typ.ArrayType]
        | type[typ.MappingType]
        | type[typ.Typed]
        | type[typ.AnyType]
        ),
    ) -> (
        typ.VariadicArrayType
        | typ.ArrayType
        | typ.MappingType
        | typ.Typed
        | typ.AnyType
        | enm.ParseErrorRef
        ):
    """
    Try to recursively parse python `tp` from `value`.

    ---

    Value should either be an instance of `tp` or a valid, serialized \
    representation of that `tp` (JSON string or otherwise).

    Returns `enm.ParseErrorRef` if no valid type could be parsed, \
    allowing for downstream validation instead of immediately raising \
    an exception within this function.

    """

    valid_types = typ.utl.check.expand_types(tp)
    if len(valid_types) > 1:
        parsed_value_or_err_ref = parse(value, valid_types[0])
        for dtype_candidate in valid_types[1:]:
            if not isinstance(parsed_value_or_err_ref, enm.ParseErrorRef):
                break
            else:  # pragma: no cover
                parsed_value_or_err_ref = parse(value, dtype_candidate)
        return parsed_value_or_err_ref
    elif isinstance(value, str):
        if (
            typ.utl.check.is_array_type(tp)
            or typ.utl.check.is_variadic_array_type(tp)
            ):
            deserialized_as_list = try_parse_json(value)
            if not isinstance(deserialized_as_list, enm.ParseErrorRef):
                return parse(deserialized_as_list, tp)
            else:
                return deserialized_as_list
        elif typ.utl.check.is_mapping_type(tp):
            deserialized_as_dict = try_parse_json(value)
            if not isinstance(deserialized_as_dict, enm.ParseErrorRef):
                return parse(deserialized_as_dict, tp)
            else:
                return deserialized_as_dict
        else:
            return try_decode(value, tp)
    elif typ.utl.check.is_typed(tp):
        if typ.utl.check.is_object(tp):
            tp_annotations = {
                k: typ.utl.check.get_args(v)[0]  # Expand Field[Any] --> Any
                for k, v
                in typ.utl.hint.collect_annotations(tp).items()
                }
        else:
            tp_annotations = typ.utl.hint.collect_annotations(tp)
        if typ.utl.check.is_serialized_mapping(value):
            tp_dict: dict[str, lib.t.Any] = {}
            for k, val in value.items():
                if (
                    isinstance(k, str)
                    and (
                        ckey := strings.utl.cname_for(
                            k,
                            tuple(tp_annotations)
                            )
                        )
                    ):
                    tp_val = parse(val, tp_annotations[ckey])
                    if isinstance(tp_val, enm.ParseErrorRef):
                        return enm.ParseErrorRef.invalid_map_decode
                    tp_dict[ckey] = tp_val
                else:
                    return enm.ParseErrorRef.invalid_keys_decode
            return tp(**tp_dict)
        else:  # pragma: no cover
            return try_decode(value, tp)
    elif (generics := typ.utl.check.get_type_args(tp)):
        if (typ.utl.check.is_variadic_array_type(tp)):
            if not typ.utl.check.is_array(value):
                return enm.ParseErrorRef.value_decode
            elif typ.utl.check.is_ellipsis(generics[-1]):
                parsed_variadic_unknown_len = [
                    parse(v, generics[0])
                    for v
                    in value
                    ]
                if any(
                    (
                        isinstance(p, enm.ParseErrorRef)
                        or not isinstance(
                            p,
                            typ.utl.check.get_checkable_types(generics[0])
                            )
                        )
                    for p
                    in parsed_variadic_unknown_len
                    ):
                    return enm.ParseErrorRef.invalid_arr_decode
                else:
                    return try_decode(parsed_variadic_unknown_len, tp)
            elif len(value) == len(generics):
                parsed_variadic_known_len = [
                    parse(v, generics[i])
                    for i, v
                    in enumerate(value)
                    ]
                if any(
                    (
                        isinstance(p, enm.ParseErrorRef)
                        or not isinstance(
                            p,
                            typ.utl.check.get_checkable_types(generics[i])
                            )
                        )
                    for i, p
                    in enumerate(parsed_variadic_known_len)
                    ):
                    return enm.ParseErrorRef.invalid_arr_decode
                else:
                    return try_decode(parsed_variadic_known_len, tp)
            else:
                return enm.ParseErrorRef.invalid_arr_len
        elif typ.utl.check.is_array_type(tp):
            if typ.utl.check.is_array(value):
                parsed_array = [
                    parse(v, generics[0])
                    for v
                    in value
                    ]
                if any(
                    (
                        isinstance(p, enm.ParseErrorRef)
                        or not isinstance(
                            p,
                            typ.utl.check.get_checkable_types(generics[0])
                            )
                        )
                    for p
                    in parsed_array
                    ):
                    return enm.ParseErrorRef.invalid_arr_decode
                else:
                    return try_decode(parsed_array, tp)
            else:
                return try_decode(value, tp)
        elif typ.utl.check.is_mapping_type(tp):
            if (
                typ.utl.check.is_serialized_mapping(value)
                or typ.utl.check.is_mapping(value)
                ):
                if len(generics) == 2:
                    key_type, value_type = generics
                    parsed_map = {
                        parse(k, key_type): parse(v, value_type)
                        for k, v
                        in value.items()
                        }
                    if any(
                        isinstance(k, enm.ParseErrorRef)
                        for k
                        in parsed_map.keys()
                        ):
                        return enm.ParseErrorRef.invalid_keys_decode
                    elif any(
                        isinstance(v, enm.ParseErrorRef)
                        for v
                        in parsed_map.values()
                        ):
                        return enm.ParseErrorRef.invalid_values_decode
                    else:
                        return try_decode(parsed_map, tp)
                else:
                    return enm.ParseErrorRef.invalid_map_decode
            else:
                return try_decode(value, tp)
        else:  # pragma: no cover
            return try_decode(value, tp)
    else:
        return try_decode(value, tp)



"""CLI objects."""

__all__ = (
    'parsers',
    'root_parser',
    )

from .. import __version__

from . import cfg
from . import lib


class Constants(cfg.Constants):
    """Constant values specific to CLI objs."""


root_parser = lib.argparse.ArgumentParser(
    description='root_parser',
    formatter_class=lib.argparse.ArgumentDefaultsHelpFormatter,
    prog='fqr',
    )
root_parser.add_argument(
    '--version',
    '-v',
    action='version',
    version=__version__,
    )

parsers = root_parser.add_subparsers(
    title='module',
    required=True,
    help='specify a module to access its commands'
    )



"""Core enumerations."""

__all__ = (
    'Boolean',
    'NoneAlias',
    )

from . import lib


class Boolean(lib.enum.Enum):
    """Boolean Enumeration."""

    true  = True
    false = False


class NoneAlias(lib.enum.Enum):
    """Nones Enumeration."""

    null = None
    none = None
    nan  = None



"""
Objects Overview
================

**Author:** dan@1howardcapital.com

**Summary:** Objects module including `Object` and `Field`.

---

Usage
-----

```py
import fqr

```

"""

from . import cfg
from . import enm
from . import exc
from . import fields
from . import lib
from . import metas
from . import objs
from . import queries
from . import typ
from . import utl

__all__ = (
    'cfg',
    'enm',
    'exc',
    'fields',
    'lib',
    'metas',
    'objs',
    'queries',
    'typ',
    'utl',
    'Field',
    'Object'
    )

from . fields import Field
from . objs import Object



"""Loggers imports."""

from .. import core

__all__ = (
    'logging',
    'time',
    'traceback',
    'warnings',
    *core.lib.__all__
    )

import logging
import time
import traceback
import warnings

from .. core . lib import *



"""Type checking utility functions."""

__all__ = (
    'get_args',
    'get_checkable_types',
    'expand_types',
    'is_array',
    'is_array_type',
    'is_bool_type',
    'is_date_type',
    'is_datetime_type',
    'is_field',
    'is_field_type',
    'is_immutable_type',
    'is_literal',
    'is_mapping',
    'is_mapping_type',
    'is_none_type',
    'is_number_type',
    'is_object',
    'is_params_type',
    'is_primitive',
    'is_serialized_mapping',
    'is_typed',
    'is_typevar',
    'is_union',
    'is_variadic_array_type',
    'is_wrapper_type',
    )

from .. import cfg
from .. import lib
from .. import obj

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from .. import typ


class Constants(cfg.Constants):
    """Constant values specific to this file."""


def get_args(tp: lib.t.Any) -> tuple[lib.t.Any, ...]:
    """Wrapper for `lib.t.get_args`."""

    return lib.t.get_args(tp)


@lib.t.overload
def get_type_args(
    tp: 'obj.SupportsParams[lib.Unpack[typ.ArgsType]]'
    ) -> 'tuple[lib.Unpack[typ.ArgsType]]': ...
@lib.t.overload
def get_type_args(tp: lib.t.Any) -> tuple[lib.t.Any, ...]: ...
def get_type_args(
    tp: 'obj.SupportsParams[lib.Unpack[typ.ArgsType]] | lib.t.Any'
    ) -> 'tuple[lib.Unpack[typ.ArgsType]] | tuple[lib.t.Any, ...]':
    """
    Get generic arguments for `type[Any]`.

    ---

    `Literals` will be returned as the types of their values.

    #### Examples

    ```py
    get_type_args(Literal[1])
    (int, )

    get_type_args(tuple[str, int])
    (str, int, )

    ```

    """

    return tuple(
        type(arg)
        if is_literal(tp)
        else arg
        for arg
        in get_args(tp)
        )


@lib.t.overload
def get_checkable_types(
    any_tp: 'type[typ.AnyType]'
    ) -> 'tuple[type[typ.AnyType] | type, ...]': ...
@lib.t.overload
def get_checkable_types(
    any_tp: lib.t.Any
    ) -> tuple[type, ...]: ...
def get_checkable_types(
    any_tp: 'type[typ.AnyType] | type[lib.t.Any] | lib.t.Any'
    ) -> 'tuple[typ.AnyType | type, ...] | tuple[type, ...]':
    """
    Get checkable origin lib.types, handling `Union` and `TypeVar` \
    expansions automatically.

    ---

    `Literals` will be returned as their values.

    `Annotated`, `ClassVar`, `Final`, and `InitVar` are expanded as \
    their parameter arguments.

    """

    checkable_types = {
        otp
        for tp
        in expand_types(any_tp)
        if isinstance((otp := lib.t.get_origin(tp) or tp), type)
        }

    return tuple(checkable_types)


@lib.t.overload
def expand_types(
    any_tp: 'type[typ.AnyType]'
    ) -> 'tuple[type[typ.AnyType] | type[lib.t.Any], ...]': ...
@lib.t.overload
def expand_types(
    any_tp: lib.t.Any
    ) -> 'tuple[type[lib.t.Any], ...]': ...
def expand_types(
    any_tp: 'type[typ.AnyType] | lib.t.Any'
    ) -> lib.t.Union[
        'tuple[type[typ.AnyType] | type[lib.t.Any], ...]',
        tuple[type[lib.t.Any], ...]
        ]:
    """
    Recursively get valid subtypes into flattened `tuple` from a \
    passed `type`, `Union`, or `TypeVar`.

    ---

    `Literals` will be returned as their values.

    `Annotated`, `ClassVar`, `Final`, and `InitVar` are expanded as \
    their parameter arguments.

    """

    if is_union(any_tp) or is_wrapper_type(any_tp):
        return tuple(
            tp
            for tp
            in set(
                lib.itertools.chain.from_iterable(
                    expand_types(sub_tp)
                    for sub_tp
                    in get_type_args(any_tp)
                    )
                )
            )
    elif is_typevar(any_tp):
        if any_tp.__constraints__:
            return tuple(
                tp
                for tp
                in set(
                    lib.itertools.chain.from_iterable(
                        expand_types(sub_tp)
                        for sub_tp
                        in any_tp.__constraints__
                        )
                    )
                )
        elif any_tp.__bound__:
            return expand_types(any_tp.__bound__)
        else:
            return (object, )
    elif is_literal(any_tp):
        return (type(get_args(any_tp)[0]), )
    else:
        return (any_tp, )


@lib.t.overload
def is_params_type(
    tp: 'obj.SupportsParams[lib.Unpack[typ.ArgsType]]',
    ) -> lib.t.TypeGuard[
        'obj.SupportsParams[lib.Unpack[typ.ArgsType]]'
        ]: ...
@lib.t.overload
def is_params_type(
    tp: lib.t.Any
    ) -> lib.t.TypeGuard[
        'obj.SupportsParams[lib.Unpack[tuple[lib.t.Any, ...]]]'
        ]: ...
def is_params_type(
    tp: 'obj.SupportsParams[lib.Unpack[typ.ArgsType]] | lib.t.Any'
    ) -> lib.t.TypeGuard[
        lib.t.Union[
            'obj.SupportsParams[lib.Unpack[typ.ArgsType]]',
            'obj.SupportsParams[lib.Unpack[tuple[lib.t.Any, ...]]]'
            ]
        ]:
    """Return `True` if `tp` has type args."""

    return bool(get_args(tp))


def is_typevar(
    obj: lib.t.Any
    ) -> lib.t.TypeGuard[lib.t.TypeVar]:
    """Return `True` if obj is a `TypeVar`."""

    return isinstance(obj, lib.t.TypeVar)


def is_union(
    obj: lib.t.Any
    ) -> lib.t.TypeGuard[lib.lib.types.UnionType]:
    """Return `True` if obj is a `UnionType`."""

    from .. import typ

    return isinstance(
        obj,
        (
            typ.OptionalGenericAlias,
            typ.UnionGenericAlias,
            lib.types.UnionType
            )
        )


@lib.t.overload
def is_bool_type(
    tp: 'type[typ.AnyType]'
    ) -> 'lib.t.TypeGuard[type[typ.AnyType]]': ...
@lib.t.overload
def is_bool_type(
    tp: type[lib.t.Any] | lib.t.Any
    ) -> lib.t.TypeGuard[type[bool]]: ...
def is_bool_type(
    tp: 'type[typ.AnyType] | type[lib.t.Any] | lib.t.Any'
    ) -> 'lib.t.TypeGuard[type[bool] | type[typ.AnyType]]':
    """Return `True` if `tp` is `type[bool]`."""

    otps = get_checkable_types(tp)

    if otps:
        return issubclass(otps[0], bool)
    else:
        return False


@lib.t.overload
def is_number_type(
    tp: 'type[typ.NumberType]'
    ) -> 'lib.t.TypeGuard[type[typ.NumberType]]': ...
@lib.t.overload
def is_number_type(
    tp: type[lib.t.Any] | lib.t.Any
    ) -> lib.t.TypeGuard[type[lib.numbers.Number]]: ...
def is_number_type(
    tp: 'type[typ.NumberType] | type[lib.t.Any] | lib.t.Any'
    ) -> 'lib.t.TypeGuard[type[lib.numbers.Number] | type[typ.NumberType]]':
    """Return `True` if tp is `numbers.Number`."""

    otps = get_checkable_types(tp)

    if otps:
        return issubclass(tp, get_checkable_types(lib.numbers.Number))
    else:
        return False


def is_ellipsis(
    tp: type[lib.t.Any] | lib.t.Any
    ) -> lib.t.TypeGuard[lib.types.EllipsisType]:
    """Return `True` if `tp` is `[...]`."""

    otps = get_checkable_types(type(tp))

    if otps:
        return issubclass(otps[0], lib.types.EllipsisType)
    else:  # pragma: no cover
        return False


def is_literal(
    tp: type[lib.t.Any] | lib.t.Any
    ) -> 'lib.t.TypeGuard[typ.Literal]':
    """Return `True` if `tp` is a `Literal`."""

    otp = lib.t.get_origin(tp) or tp

    return getattr(otp, '__name__', '') == 'Literal'


def is_date_type(
    tp: type[lib.t.Any] | lib.t.Any,
    ) -> lib.t.TypeGuard[type[lib.datetime.date]]:
    """Return `True` if `tp` is `datetime.date`."""

    otps = get_checkable_types(tp)

    if otps:
        return issubclass(otps[0], lib.datetime.date)
    else:
        return False


def is_datetime_type(
    tp: type[lib.t.Any] | lib.t.Any,
    ) -> lib.t.TypeGuard[type[lib.datetime.datetime]]:
    """Return `True` if `tp` is `datetime.datetime`."""

    otps = get_checkable_types(tp)

    if otps:
        return issubclass(otps[0], lib.datetime.datetime)
    else:
        return False


def is_none_type(
    tp: type[lib.t.Any] | lib.t.Any,
    ) -> 'lib.t.TypeGuard[type[typ.NoneType]]':
    """Return `True` if `tp` is `NoneType`."""

    otps = get_checkable_types(tp)

    if otps:
        return issubclass(otps[0], None.__class__)
    else:
        return False


def is_primitive(
    obj: lib.t.Any
    ) -> 'lib.t.TypeGuard[typ.Primitive]':
    """Return `True` if `obj` is a `Primitive`."""

    from .. import typ

    return isinstance(obj, get_checkable_types(typ.Primitive))


def is_serialized_mapping(
    obj: lib.t.Any | type[lib.t.Any]
    ) -> 'lib.t.TypeGuard[typ.Mapping[typ.Primitive, typ.Serial]]':
    """
    Return `True` if `obj` is `MappingProto[typ.Primitive, typ.Serial]`.

    """

    from .. import typ

    return (
        isinstance(obj, lib.t.Mapping)
        and all(
            (
                is_primitive(k)
                and isinstance(v, get_checkable_types(typ.Serial))
                )
            for k, v
            in obj.items()
            )
        )


def is_mapping(
    obj: 'typ.MappingType | lib.t.Any'
    ) -> lib.t.TypeGuard[
        'typ.MappingType | typ.Mapping[lib.t.Any, lib.t.Any]'
        ]:
    """Return `True` if `obj` is `Mapping[lib.t.Any, lib.t.Any]`."""

    return isinstance(obj, lib.t.Mapping)


def is_mapping_type(
    tp: 'type[typ.MappingType] | type[lib.t.Any] | lib.t.Any'
    ) -> 'lib.t.TypeGuard[type[typ.MappingType]]':
    """Return `True` if `tp` is `type[Mapping[lib.t.Any, lib.t.Any]]`."""

    otps = get_checkable_types(tp)

    if otps:
        return issubclass(otps[0], lib.t.Mapping)
    else:
        return False


@lib.t.overload
def is_array(
    obj: 'typ.Array[typ.AnyType]',
    ) -> 'lib.t.TypeGuard[typ.Array[typ.AnyType]]': ...
@lib.t.overload
def is_array(
    obj: lib.t.Any,
    ) -> 'lib.t.TypeGuard[typ.Array[lib.t.Any]]': ...
def is_array(
    obj: 'typ.Array[typ.AnyType | lib.t.Any] | lib.t.Any'
    ) -> 'lib.t.TypeGuard[typ.Array[typ.AnyType | lib.t.Any]]':
    """Return `True` if `obj` is `Array[lib.t.Any]`."""

    return (
        isinstance(obj, lib.t.Collection)
        and not isinstance(obj, (str, lib.t.Mapping))
        )


def is_object(
    obj_: 'obj.ObjectLike | type[lib.t.Any] | lib.t.Any'
    ) -> lib.t.TypeGuard['obj.ObjectLike']:
    """Return `True` if `obj_` is an `Object`."""

    if isinstance(obj_, type):
        otp = lib.t.get_origin(obj_) or obj_
    else:
        otp = type(obj_)

    from .... import objects

    return issubclass(otp, objects.Object)


def is_field(
    obj_: 'typ.AnyField[typ.AnyType] | lib.t.Any'
    ) -> lib.t.TypeGuard[
        'typ.AnyField[typ.AnyType] | typ.AnyField[lib.t.Any]'
        ]:
    """Return `True` if `obj_` typ AnyField[Any]`."""

    obj_tp = obj_ if isinstance(obj_, type) else type(obj_)

    return is_field_type(obj_tp)


def is_field_type(
    tp: 'type[typ.AnyField[typ.AnyType]] | lib.t.Any'
    ) -> lib.t.TypeGuard[
        'type[typ.AnyField[typ.AnyType]] | type[typ.AnyField[lib.t.Any]]'
        ]:
    """Return `True` if `tp` is `typ[AnyField[Any]]`."""

    if isinstance(tp, lib.t.ForwardRef):
        return bool(obj.FieldPattern.match(tp.__forward_arg__))
    elif isinstance(tp, str):
        return bool(obj.FieldPattern.match(tp))

    otp = lib.t.get_origin(tp) or tp

    from .... import objects

    return (
        getattr(otp, '__name__', '') == 'Field'
        and not issubclass(otp, objects.typ.Field)
        )


def is_wrapper_type(
    tp: 'typ.Wrapper | type[lib.t.Any] | lib.t.Any'
    ) -> lib.t.TypeGuard['typ.Wrapper']:
    """Return `True` if `tp` is `Annotated | ClassVar | Final | InitVar`."""

    if isinstance(tp, lib.t.ForwardRef):
        return bool(obj.WrapperPattern.match(tp.__forward_arg__))
    elif isinstance(tp, str):  # pragma: no cover
        return bool(obj.WrapperPattern.match(tp))

    otp = lib.t.get_origin(tp) or tp

    return (
        getattr(otp, '__name__', '')
        in {'Annotated', 'ClassVar', 'Final', 'InitVar'}
        )


def is_typed(
    any: 'type[typ.Typed] | type[lib.t.Any] | lib.t.Any'
    ) -> 'lib.t.TypeGuard[type[typ.Typed]]':
    """Return `True` if `any` is type-hinted."""

    return (
        getattr(any, '__annotations__', False) is not False
        and not isinstance(any, lib.t.ForwardRef)
        and not isinstance(
            any,
            (
                lib.types.FunctionType,
                lib.types.MethodType
                )
            )
        )


def is_array_type(
    tp: 'type[typ.ArrayType] | type[lib.t.Any] | lib.t.Any'
    ) -> 'lib.t.TypeGuard[type[typ.ArrayType]]':
    """Return `True` if `tp` is `type[Array[lib.t.Any]]`."""

    otps = get_checkable_types(tp)

    if otps:
        return (
            issubclass(otps[0], lib.t.Collection)
            and not issubclass(otps[0], (str, lib.t.Mapping))
            )
    else:
        return False


def is_variadic_array_type(
    tp: 'type[typ.VariadicArrayType] | type[lib.t.Any] | lib.t.Any'
    ) -> 'lib.t.TypeGuard[type[typ.VariadicArrayType]]':
    """Return `True` if `tp` is `type[VariadicArray]`."""

    otps = get_checkable_types(tp)

    if otps:
        return issubclass(otps[0], tuple)
    else:
        return False


def is_immutable_type(
    tp: type[lib.t.Any],
    ) -> 'lib.t.TypeGuard[type[typ.Immutable]]':
    """Return `True` if tp is an immutable standardlib type."""

    from .. import typ

    for sub_tp in expand_types(tp):
        if all(
            issubclass(sub_otp, get_checkable_types(typ.Immutable))
            for sub_otp
            in get_checkable_types(sub_tp)
            ):
            if is_params_type(sub_tp):
                if all(
                    is_immutable_type(arg_tp)
                    or is_ellipsis(arg_tp)
                    for arg_tp
                    in get_type_args(sub_tp)
                    ):
                    return True
            else:
                return True

    return False



"""Module utils unit tests."""

import unittest

import fqr

from . import cfg


class Constants(cfg.Constants):
    """Constant values specific to unit tests in this file."""

    VALID_CAMEL_STRING_EXAMPLES = (
        'upperSnakeCaseString100SureFift337aGood',
        '_privateValidCamelCaseString'
        )
    VALID_SNAKE_STRING_EXAMPLES = (
        'upper_snake_case_string_100_sure_fift_337a_good',
        '_snake_case_string_'
        )
    BIG_STR_DICT = {
        'api-key': VALID_CAMEL_STRING_EXAMPLES[0] * 30,
        'regular': VALID_SNAKE_STRING_EXAMPLES[0] * 30,
        }


class TestUtils(unittest.TestCase):
    """Fixture for testing."""

    def test_01_camel_case_check(self):
        """Test camelCase validation."""

        self.assertFalse(
            fqr.core.strings.utl.isCamelCaseString(
                Constants.VALID_CAMEL_STRING_EXAMPLES[-1]
                )
            )

    def test_02_case_conversion(self):
        """Test case conversion."""

        self.assertTrue(
            fqr.core.strings.utl.is_snake_case_iterable(
                [
                    fqr.core.strings.utl.camel_case_to_snake_case(string)
                    for string
                    in Constants.VALID_CAMEL_STRING_EXAMPLES
                    ]
                )
            )

    def test_03_str_validation(self):
        """Test case validation asserts `type[str]`."""

        self.assertRaises(
            TypeError,
            lambda: fqr.core.strings.utl.validate_casing(
                123,
                fqr.core.strings.enm.SupportedCasing.snake_case.value
                )
            )

    def test_04_case_validation(self):
        """Test case validation asserts snake casing."""

        self.assertRaises(
            fqr.core.strings.exc.StringCasingError,
            lambda: fqr.core.strings.utl.validate_casing(
                Constants.INVALID_STRING_CASING_EXAMPLE,
                fqr.core.strings.enm.SupportedCasing.snake_case.value
                )
            )

    def test_05_case_validation(self):
        """Test case validation asserts camel casing."""

        self.assertRaises(
            fqr.core.strings.exc.StringCasingError,
            lambda: fqr.core.strings.utl.validate_casing(
                Constants.INVALID_STRING_CASING_EXAMPLE,
                fqr.core.strings.enm.SupportedCasing.camelCase.value
                )
            )

    def test_06_snake_iterable_check(self):
        """Test snake_case validation."""

        self.assertTrue(
            fqr.core.strings.utl.is_snake_case_iterable(
                Constants.VALID_SNAKE_STRING_EXAMPLES
                )
            )

    def test_07_camel_iterable_check(self):
        """Test camelCase validation."""

        self.assertTrue(
            fqr.core.strings.utl.isCamelCaseIterable(
                Constants.VALID_CAMEL_STRING_EXAMPLES
                )
            )

    def test_08_get_cname(self):
        """Test cname retrieval."""

        self.assertEqual(
            fqr.core.strings.utl.cname_for(
                Constants.VALID_CAMEL_STRING_EXAMPLES[-1].strip('_'),
                Constants.VALID_CAMEL_STRING_EXAMPLES
                ),
                Constants.VALID_CAMEL_STRING_EXAMPLES[-1]
            )

    def test_09_get_cname(self):
        """Test cname retrieval for a snake_case string."""

        self.assertEqual(
            fqr.core.strings.utl.cname_for(
                Constants.VALID_SNAKE_STRING_EXAMPLES[0],
                Constants.VALID_CAMEL_STRING_EXAMPLES
                ),
                Constants.VALID_CAMEL_STRING_EXAMPLES[0]
            )

    def test_10_get_cname(self):
        """Test cname retrieval negative case returns None."""

        self.assertIsNone(
            fqr.core.strings.utl.cname_for(
                Constants.INVALID_STRING_CASING_EXAMPLE,
                Constants.VALID_CAMEL_STRING_EXAMPLES
                )
            )

    def test_11_validation_success(self):
        """Test validation succeeds."""

        self.assertIsNone(
            fqr.core.strings.utl.validate_casing(
                Constants.VALID_SNAKE_STRING_EXAMPLES[0],
                fqr.core.strings.enm.SupportedCasing.snake_case.value
                )
            )

    def test_12_field_serialization(self):
        """Test field serialization."""

        field = fqr.Field(name='test', default='testing', type_=str)
        self.assertEqual(
            repr(fqr.objects.typ.Field(field, field.type_)),
            fqr.core.strings.utl.convert_for_repr(field)
            )

    def test_13_fn_serialization(self):
        """Test fn serialization."""

        def fn() -> str:
            return 'str'

        self.assertEqual(
            fqr.core.strings.utl.convert_for_repr(
                fn.__annotations__['return'],
                ),
            fqr.core.strings.utl.convert_for_repr(fn)
            )

    def test_14_big_str_redaction(self):
        """Test big_str redaction."""

        self.assertEqual(
            fqr.core.strings.utl.convert_for_repr(
                Constants.BIG_STR_DICT,
                )['api-key'],
            fqr.core.strings.utl.redact_key_value_pair('api-key', 'anything')
            )

    def test_15_big_str_serialization(self):
        """Test big_str serialization."""

        self.assertEqual(
            fqr.core.strings.utl.convert_for_repr(
                Constants.BIG_STR_DICT['regular']
                ),
            [
                Constants.M_LINE_TOKEN,
                *fqr.core.strings.obj.StringWrapper.wrap(
                    Constants.BIG_STR_DICT['regular']
                    )
                ]
            )
